/////////////////////////////////////////////////////////////////////////////
// Name:        TAS3DCanvas.cpp
// Purpose:     Defines the canvas used for 3-D rendering
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"

#ifdef __GNUG__
#pragma implementation
#pragma interface
#endif

// For compilers that support precompilation, includes "wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#if !wxUSE_GLCANVAS
#error "OpenGL required: set wxUSE_GLCANVAS to 1 and rebuild the wxWidgets libraries"
#endif

#ifdef __WXMAC__
  #include <OpenGL/glu.h>
  #include <GLUT/glut.h>
#else
  #include <GL/glu.h>
  #include <GL/glut.h>
#endif /* __WXMAC__ */
#include <math.h>
#include "TASPanel.h"
#include "TAS3DCanvas.h"
#include "TASTexFont.h"
#include "TASTiles.h"
#include "TASDockers.h"
#include "TASApp.h"
#include "TASDebug.h"


#ifndef max
#define max(a, b) (((a) > (b)) ? (a) : (b))
#endif
#define glRGB(x, y, z)	glColor3ub((GLubyte)x, (GLubyte)y, (GLubyte)z)
#define RADS_IN_CIRCLE 6.28318531f

#define TAS_3D_MOUSE_LCLICK   0
#define TAS_3D_MOUSE_LDCLICK  1
#define TAS_3D_MOUSE_OVER     2

// Xt replacement functions
char* XtMalloc(size_t size)
{
  char* ptr = (char*)malloc(size);
  if (!ptr)
  {
    perror("malloc in XtMalloc()");
    exit(1);
  }
  return ptr;
}

char* XtNewString(const char* string)
{
  char* ptr = (char*)malloc(strlen(string)+1);
  if (!ptr)
  {
    perror("malloc in XtNewString()");
    exit(1);
  }
  ptr[strlen(string)] = 0;
  return strcpy(ptr, string);
}

char* XtCalloc(size_t num, size_t size)
{
  char* ptr = (char*)calloc(num, size);
  if (!ptr) {
    perror("malloc in XtCalloc()");
    exit(1);
  }
  return ptr;
}

char* XtRealloc(char* ptrIn, size_t size)
{
  char* ptr = (char*)realloc(ptrIn, size);
  if (!ptr) {
    perror("realloc in XtRealloc()");
    exit(1);
  }
  return ptr;
}

void XtFree(void* ptr)
{
  if (ptr)
  {
    free(ptr);
    ptr = NULL;
  }
}

void XtError(char* err_msg)
{
  perror(err_msg);
  exit(1);
}

bool usePolygonOffset = true;

void TAS3DCanvas::CubeSide(GLfloat size, int mode, bool bWireframe, wxColour* pColor)
{
  GLfloat fWireWidth = 0.0f;//0.05f;
  glRGB(pColor->Red(), pColor->Green(), pColor->Blue());

  if (!bWireframe) {
    glPolygonMode(GL_FRONT, GL_FILL);
    glRectf(-((size/2.0)-fWireWidth), -((size/2.0)-fWireWidth), ((size/2.0)-fWireWidth), ((size/2.0)-fWireWidth));
  }
}

void TAS3DCanvas::CubeSideWithText(GLfloat size, char *text, int len, int mode,
                                     int rotateCount, wxColour* pColor)
{
  int w, a, d;

//  CubeSide(size, mode, false, pColor);

  //if (!IsInOverviewWin() && m_bShowSymbols) {
    glPushMatrix();
      glEnable(GL_TEXTURE_2D);
      //  alphaModeSet();
      glEnable(GL_ALPHA_TEST);
      glAlphaFunc(GL_GEQUAL, 0.5);
  
      if (usePolygonOffset) {
    #if defined(GL_EXT_polygon_offset) || defined(GL_VERSION_1_1)
        glEnable(GL_POLYGON_OFFSET_FILL);
        glPolygonOffset(0.0, -1.0f);
    #endif
      }
  
      txfGetStringMetrics(m_pTexFont, text, len, &w, &a, &d);

      glScalef(1.0/w, 1.0/w, 1.0/w);

      glRGB(g_pContext->m_colDefaultText.Red(), g_pContext->m_colDefaultText.Green(), g_pContext->m_colDefaultText.Blue());

      glPushMatrix();
        glRotatef((rotateCount * 90.0f), 0.0f, 0.0f, 1.0f);
        glTranslatef(-w/2.0, d-(a+d)/2.0, 0.0);
//        txfRenderFancyString(m_pTexFont, text, len);
        txfRenderString(m_pTexFont, text, len);
      glPopMatrix();

      //  Draw underline to make symbol orientation obvious
      glTranslatef(-w/2.0, d-(a+d)/2.0, 0.0);
//      txfRenderFancyString(m_pTexFont, "_", 1);
      txfRenderString(m_pTexFont, "_", 1);
  
      if (usePolygonOffset) {
    #if defined(GL_EXT_polygon_offset) || defined(GL_VERSION_1_1)
        glDisable(GL_POLYGON_OFFSET_FILL);
    #endif
      }

      glDisable(GL_TEXTURE_2D);
      glDisable(GL_ALPHA_TEST);
  
    glPopMatrix();
  //}
}

//MJP: reduce this to just draw individual chars to improve performance?
void TAS3DCanvas::TextedCube(GLfloat size, char* text, int mode, int rotateCount, wxColour* pColor)
{
  glRGB(pColor->Red(), pColor->Green(), pColor->Blue());
  glutSolidCube(size);

  glPushMatrix();
    glTranslatef(0.0, 0.0, size/2.0);
    CubeSideWithText(size, text, (int)strlen(text), mode, rotateCount, pColor);
  glPopMatrix();
  
  glPushMatrix();
    glRotatef(90.0, 0, 1, 0);
    glTranslatef(0.0, 0.0, size/2.0);
    CubeSideWithText(size, text, (int)strlen(text), mode, rotateCount, pColor);
  glPopMatrix();
  
  glPushMatrix();
    glRotatef(180.0, 0, 1, 0);
    glTranslatef(0.0, 0.0, size/2.0);
    CubeSideWithText(size, text, (int)strlen(text), mode, rotateCount, pColor);
  glPopMatrix();
  
  glPushMatrix();
    glRotatef(270.0, 0, 1, 0);
    glTranslatef(0.0, 0.0, size/2.0);
    CubeSideWithText(size, text, (int)strlen(text), mode, rotateCount, pColor);
  glPopMatrix();
  
  glPushMatrix();
    glRotatef(90.0, 1, 0, 0);
    glTranslatef(0.0, 0.0, size/2.0);
    CubeSideWithText(size, text, (int)strlen(text), mode, rotateCount, pColor);
  glPopMatrix();
  
  glPushMatrix();
    glRotatef(270.0, 1, 0, 0);
    glTranslatef(0.0, 0.0, size/2.0);
    CubeSideWithText(size, text, (int)strlen(text), mode, rotateCount, pColor);
  glPopMatrix();

  if (mode == TAS_3D_CUBE_SELECTED) {
    glRGB(255, 0, 0);
  } else if (mode == TAS_3D_CUBE_HIGHLIGHTED) {
    glRGB(255, 255, 255);
  } else {
    glColor3ub(m_colWire.Red(), m_colWire.Green(), m_colWire.Blue());
  }

  glutWireCube(size);
}

void TAS3DCanvas::WireframeCube(GLfloat size, int mode)
{
  glRGB(m_colWire.Red(), m_colWire.Green(), m_colWire.Blue());
  glutWireCube(size);
}


TASOGLCanvas::TASOGLCanvas(wxScrolledWindow* parent, TAS3DPanel* pMainPanel,
                               TASSimulatorFrame* parentFrame, wxWindowID id, const wxPoint& pos,
                               const wxSize& size, long style, const wxString& name)
                : wxGLCanvas(parent, id, pos, size, style|wxFULL_REPAINT_ON_RESIZE, name)
{
}

BEGIN_EVENT_TABLE(TAS3DCanvas, wxGLCanvas)
  EVT_SIZE(TAS3DCanvas::OnSize)
  EVT_PAINT(TAS3DCanvas::OnPaint)
  EVT_ERASE_BACKGROUND(TAS3DCanvas::OnEraseBackground)
  EVT_MOUSE_EVENTS(TAS3DCanvas::OnMouse)
  EVT_KEY_DOWN(TAS3DCanvas::OnKeyboard)
END_EVENT_TABLE()

TAS3DCanvas::TAS3DCanvas(wxScrolledWindow* parent, TAS3DPanel* pMainPanel, TASSimulatorFrame* parentFrame, wxWindowID id,
                             const wxPoint& pos, const wxSize& size, long style, const wxString& name)
                : TASOGLCanvas(parent, pMainPanel, parentFrame, id, pos, size, style, name)
{
  m_parent         = parent;
  m_parentFrame    = parentFrame;
  m_pSpace         = NULL;
  m_bDoInit        = true;
  m_bShowMouseOver = true;
  m_bShowSymbols   = true;
  m_iViewType      = TAS_3D_VIEW_PERSPECTIVE;
  m_fBeginX        = 0.0f;
  m_fBeginY        = 0.0f;
  m_fFovy          = 75.0f;
  m_fScaleFact     = 100.0f;
  m_fNear          = 1.0f;
  m_fFar           = 900.0f;
  m_fCubeSize      = 2.0f;
  m_screenMinX     = 9999999;
  m_screenMinY     = 9999999;
  m_screenMaxX     = 0;
  m_screenMaxY     = 0;
  m_fAspect        = (GLfloat)size.GetWidth()/(GLfloat)size.GetHeight();
  m_bIsOverview    = false;

  SetMainPanel(pMainPanel);

  trackball(m_fQuat, 0.0f, 0.0f, 0.0f, 0.0f);

  m_colWire.InitFromName(_T("Grey"));
}

TAS3DCanvas::~TAS3DCanvas()
{
}

bool TAS3DCanvas::IsValid() {
  return (!m_bDoInit);
}

void TAS3DCanvas::SetMainPanel(TAS3DPanel* pPanel)
{
  m_pMainPanel = pPanel;
  if (m_parent == m_pMainPanel) {
    m_bIsOverview = false;
  } else {
    m_bIsOverview = true;
  }
}

void TAS3DCanvas::RenderTile(const TASCoords& coords, GLfloat* pxHalf, GLfloat* pyHalf, GLfloat* pzHalf,
                             TASTileType* pTileType, bool* pbSelectedCellShown)
{
  int cubeMode;
  GLuint index;
  GLfloat xLoc, yLoc, zLoc;

  SetCurrent();

  xLoc = -(*pxHalf) + (coords.x * m_fCubeSize);
  yLoc = -(*pyHalf) + (coords.y * m_fCubeSize);
  zLoc = -(*pzHalf) + (coords.z * m_fCubeSize);
  
  glPushMatrix();
  
    cubeMode = TAS_3D_CUBE_NORMAL;

    glTranslatef(xLoc, yLoc, zLoc);

    if (pTileType->IsValid()) {

      if (!IsInOverviewWin()) {
        //  Calculate index number of tile to use as unique name
        index = m_pSpace->Index(coords);
        glLoadName(index);
      
        if (m_pMainPanel->m_coordsSelCell == coords) {
          cubeMode = TAS_3D_CUBE_SELECTED;
        } else if ((coords.x == m_pMainPanel->m_coordsCurrCell.x) &&
                   (coords.y == m_pMainPanel->m_coordsCurrCell.y) &&
                   (coords.z == m_pMainPanel->m_coordsCurrCell.z))
        {
          cubeMode = TAS_3D_CUBE_HIGHLIGHTED;
        } 
      }

      wxColour colorTile = g_pContext->m_colDefaultTile;
      if (pTileType->m_colTile != wxNullColour) {
        colorTile = pTileType->m_colTile;
      }

      //  Draw cube sides with text
  //    TextedCube(m_fCubeSize, (char*)pTileType->m_sName.c_str(), cubeMode, 0, &colorTile);
      int rotateCount = 0;
      char* text = (char*)pTileType->m_sName.c_str();
      glRGB(colorTile.Red(), colorTile.Green(), colorTile.Blue());
      glutSolidCube(m_fCubeSize);

      if (!IsInOverviewWin() && m_bShowSymbols) {
        glPushMatrix();
          glTranslatef(0.0, 0.0, m_fCubeSize/2.0);
          CubeSideWithText(m_fCubeSize, text, (int)strlen(text), cubeMode, rotateCount, &colorTile);
        glPopMatrix();
        
        glPushMatrix();
          glRotatef(90.0, 0, 1, 0);
          glTranslatef(0.0, 0.0, m_fCubeSize/2.0);
          CubeSideWithText(m_fCubeSize, text, (int)strlen(text), cubeMode, rotateCount, &colorTile);
        glPopMatrix();
        
        glPushMatrix();
          glRotatef(180.0, 0, 1, 0);
          glTranslatef(0.0, 0.0, m_fCubeSize/2.0);
          CubeSideWithText(m_fCubeSize, text, (int)strlen(text), cubeMode, rotateCount, &colorTile);
        glPopMatrix();
        
        glPushMatrix();
          glRotatef(270.0, 0, 1, 0);
          glTranslatef(0.0, 0.0, m_fCubeSize/2.0);
          CubeSideWithText(m_fCubeSize, text, (int)strlen(text), cubeMode, rotateCount, &colorTile);
        glPopMatrix();
        
        glPushMatrix();
          glRotatef(90.0, 1, 0, 0);
          glTranslatef(0.0, 0.0, m_fCubeSize/2.0);
          CubeSideWithText(m_fCubeSize, text, (int)strlen(text), cubeMode, rotateCount, &colorTile);
        glPopMatrix();
        
        glPushMatrix();
          glRotatef(270.0, 1, 0, 0);
          glTranslatef(0.0, 0.0, m_fCubeSize/2.0);
          CubeSideWithText(m_fCubeSize, text, (int)strlen(text), cubeMode, rotateCount, &colorTile);
        glPopMatrix();
      }

      if (cubeMode == TAS_3D_CUBE_SELECTED) {
        glRGB(0, 0, 255);
      } else if (cubeMode == TAS_3D_CUBE_HIGHLIGHTED) {
        glRGB(255, 0, 0);
      } else {
        glColor3ub(m_colWire.Red(), m_colWire.Green(), m_colWire.Blue());
      }

      glutWireCube(m_fCubeSize);

    } else if (m_pMainPanel->m_bShowFrontier) {

      //glLoadName(TAS_TILE_NOT_FOUND);
      glColor4ub(pTileType->m_colTile.Red(), pTileType->m_colTile.Green(), pTileType->m_colTile.Blue(), 50);
      glutSolidCube(m_fCubeSize - 0.1f);
/*
      glRGB(pTileType->m_colTile.Red(), pTileType->m_colTile.Green(), pTileType->m_colTile.Blue());
      glutWireCube(m_fCubeSize);
*/
    }
  
    if (cubeMode == TAS_3D_CUBE_SELECTED) {
      *pbSelectedCellShown = true;
    }

  glPopMatrix();

///////////////////////////////////////////////
/*
            glPushMatrix();
              float fvViewMatrix[16];

              //glGetFloatv(GL_MODELVIEW_MATRIX, fvViewMatrix);
              //glGetFloatv(GL_PROJECTION_MATRIX, fvViewMatrix);
        
              //glLoadMatrixf(fvViewMatrix);
        
              //glMultMatrixf();
        
              //         glMatrixMode(GL_PROJECTION);
        
              glGetFloatv(GL_PROJECTION_MATRIX, fvViewMatrix);
        
              //          fvViewMatrix[0] = xLoc;
              //          fvViewMatrix[1] = yLoc;
              //          fvViewMatrix[2] = 0;
              //          fvViewMatrix[3] = 1;
        
              //          glVertex2f(xLoc, yLoc);
              //          glMultMatrixf(fvViewMatrix);
              //          glGetFloatv(GL_PROJECTION_MATRIX, fvViewMatrix);
        
              //          GLfloat fVect[4] = { xLoc, yLoc, 0.0f, 1.0f };
              //          GLfloat fRes[4];
        
              GLfloat fX = (fvViewMatrix[0] * xLoc) + (fvViewMatrix[4] * yLoc) + (fvViewMatrix[8] * zLoc) + (fvViewMatrix[12] * 1.0f);
              GLfloat fY = (fvViewMatrix[1] * xLoc) + (fvViewMatrix[5] * yLoc) + (fvViewMatrix[9] * zLoc) + (fvViewMatrix[13] * 1.0f);
        
        
              if (fX < fLeft)    fLeft   = fX;
              if (fX > fRight)   fRight  = fX;
              if (fY < fTop)     fTop    = fY;
              if (fY > fBottom)  fBottom = fY;
        
            glPopMatrix();
*/
///////////////////////////////////////////////
}

size_t lastEpoch = -1;
GLint dispList = -1;

//#define USE_DISP_LISTS 1

bool TAS3DCanvas::RenderTileSpace(bool* pbSelectedCellShown)
{
#ifdef __WXDEBUG__
  if (!IsInOverviewWin()) {
    wxStartTimer();
  }
#endif //  __WXDEBUG__

  GLfloat xHalf, yHalf, zHalf;

  SetCurrent();

#ifdef USE_DISP_LISTS
  if (lastEpoch != (size_t)g_pTileSpace->m_epoch) {
    glNewList(dispList, GL_COMPILE);
#endif  // USE_DISP_LISTS

  wxMutexLocker lock(*g_pMutexTileSpaceGuard);

  xHalf = ((m_fCubeSize * g_pTileSpace->GetXDim()) / 2) - (m_fCubeSize / 2);
  yHalf = ((m_fCubeSize * g_pTileSpace->GetYDim()) / 2) - (m_fCubeSize / 2);
  zHalf = ((m_fCubeSize * g_pTileSpace->GetZDim()) / 2) - (m_fCubeSize / 2);
  
  GLfloat fLeft, fRight, fTop, fBottom;
  fLeft   =  99999.0f;
  fRight  = -99999.0f;
  fTop    =  99999.0f;
  fBottom = -99999.0f;

  size_t t = 0;
  TASLocationTransition* pTrans = NULL;

  TASTileType* pTTFrontier     = &m_pMainPanel->m_ttFrontierTile;
  TASTileType* pTTFrontierDead = &m_pMainPanel->m_ttFrontierDeadTile;
  if (!m_pMainPanel->m_bShowFrontier || IsInOverviewWin()) {
    pTTFrontier     = NULL;
    pTTFrontierDead = NULL;
  }

  for (size_t e=0; e<=(size_t)g_pTileSpace->m_epoch; e++) {

    t = 0;
    while (pTrans = g_pTileSpace->GetTransitionInEpoch(e, t++)) {

      TASTileType* pTT  = NULL;
      if (TASIsTileType(pTrans->m_typeEnd)) {
        pTT = g_pTileSpace->m_tileset[pTrans->m_typeEnd];
      } else if (pTrans->m_typeEnd == TAS_TILE_FRONTIER) {
        pTT = pTTFrontier;
      } else if (pTrans->m_typeEnd == TAS_TILE_FRONTIERDEAD) {
        pTT = pTTFrontierDead;
      }

      if (pTT) {
        if (m_pMainPanel->m_iViewBy == TAS_3D_VIEWBY_UNLIMITED) {
          RenderTile(pTrans->m_coords, &xHalf, &yHalf, &zHalf, pTT, pbSelectedCellShown);
        } else if (m_pMainPanel->m_iViewBy == TAS_3D_VIEWBY_RANGE) {
          if (pTrans->m_coords.x >= m_pMainPanel->m_coordsMinVis.x && pTrans->m_coords.x <= m_pMainPanel->m_coordsMaxVis.x &&
              pTrans->m_coords.y >= m_pMainPanel->m_coordsMinVis.y && pTrans->m_coords.y <= m_pMainPanel->m_coordsMaxVis.y &&
              pTrans->m_coords.z >= m_pMainPanel->m_coordsMinVis.z && pTrans->m_coords.z <= m_pMainPanel->m_coordsMaxVis.z)
          {
            RenderTile(pTrans->m_coords, &xHalf, &yHalf, &zHalf, pTT, pbSelectedCellShown);
          }
        } else if (m_pMainPanel->m_iViewBy == TAS_3D_VIEWBY_HOOD) {
          if (m_pMainPanel->CoordsInNeighborhood(pTrans->m_coords))
          {
            RenderTile(pTrans->m_coords, &xHalf, &yHalf, &zHalf, pTT, pbSelectedCellShown);
          }
        }
      }
    }
  }

  lastEpoch = (size_t)g_pTileSpace->m_epoch;

#ifdef USE_DISP_LISTS
  glEndList();
  }

  glCallList(dispList);
#endif  // USE_DISP_LISTS

#ifdef __WXDEBUG__
  if (!IsInOverviewWin()) {
    long time = wxGetElapsedTime(false);
    wxString sTime = wxString::Format(_T("Render time: %d\n"), time);
    g_pTASApp->SetSimDebugText(sTime, true, false);
  }
#endif //  __WXDEBUG__

  return true;
}

GLdouble GetCameraDistance(TASSpace* pSpace, GLfloat cubeSize)
{
  GLdouble width  = pSpace->GetXDim() * cubeSize;
  GLdouble height = pSpace->GetYDim() * cubeSize;
  GLdouble depth  = pSpace->GetZDim() * cubeSize;
  GLdouble dist   = sqrt(width * height * depth);
  return dist;
}

void TAS3DCanvas::SizerHelper()
{
  GLfloat m[4][4];

  SetCurrent();
  wxClientDC dc(this);  //  Must always be here
  
#ifndef __WXMOTIF__
  if (!GetContext()) return;
#endif

  m_pSpace = g_pTileSpace;
  int maxDim = max(m_pSpace->GetXDim(), max(m_pSpace->GetYDim(), m_pSpace->GetZDim()));

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

//MJP: this should not be automatic.  This is the place to start reworking the zooming functionality!
    //  Translate the whole scene out and into view
    GLdouble dist = GetCameraDistance(m_pSpace, m_fCubeSize);
    glTranslatef(0.0f, 0.0f, -dist/3);

    float scaledScaleFact = m_fScaleFact / 100.0f;
    glScalef(scaledScaleFact, scaledScaleFact, scaledScaleFact);

///////////////////////////////////////////////////////////
//MJP: test multiplication by projection matrix here

    // Get The Current PROJECTION Matrix From OpenGL
    GLdouble proj[16];
	  glGetDoublev(GL_PROJECTION_MATRIX, proj);

    // Get The Current MODELVIEW Matrix From OpenGL
    GLdouble model[16];
    glGetDoublev(GL_MODELVIEW_MATRIX, model);

    // Get The Current VIEWPORT From OpenGL
    GLint view[4];
    glGetIntegerv(GL_VIEWPORT, view);

    GLdouble minX, minY, maxX, maxY, winZ;

    GLint ret = gluProject(-(dist/2), -(dist/2), 0, model, proj, view, &minX, &minY, &winZ);
          ret = gluProject( (dist/2),  (dist/2), 0, model, proj, view, &maxX, &maxY, &winZ);

//  if (m_screenMinX != minX || m_screenMinY != minY || m_screenMaxX != maxX || m_screenMaxY != maxY) {
    m_screenMinX = minX;
    m_screenMinY = minY;
    m_screenMaxX = maxX;
    m_screenMaxY = maxY;
//    TASLogMsg("G", "D", "min(%4d, %4d)  max(%4d, %4d)\n", GLint(minX), GLint(view[3] - minY), GLint(maxX), GLint(view[3] - maxY));
//  }

///////////////////////////////////////////////////////////

  //  Restore the matrix state
	glPopMatrix();	// Modelview matrix
}

void TAS3DCanvas::OnPaint(wxPaintEvent& event)
{
  if (!m_pMainPanel) return;

  GLfloat m[4][4];

  SetCurrent();

  //  Must always be here
  wxPaintDC dc(this);
  
#ifndef __WXMOTIF__
  if (!GetContext()) return;
#endif

  //  Initialize OpenGL
  if (m_bDoInit)
  {
    InitGL();
    m_bDoInit = false;
  }

  //  Set background color
  GLfloat r, g, b;
  r = (float)g_pContext->m_colSpaceBkgrnd.Red() / 255.0f;
  g = (float)g_pContext->m_colSpaceBkgrnd.Green() / 255.0f;
  b = (float)g_pContext->m_colSpaceBkgrnd.Blue() / 255.0f;
  glClearColor(r, g, b, 1.0f);

  m_pSpace = g_pTileSpace;

  int maxDim = max(m_pSpace->GetXDim(), max(m_pSpace->GetYDim(), m_pSpace->GetZDim()));

  bool bSelectedCellShown = false;

  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  
  txfEstablishTexture(m_pTexFont, 1, GL_TRUE);

  //  Clear the window with current clearing color
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  build_rotmatrix(m, m_pMainPanel->m_pCanvas->m_fQuat);

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

    //  Translate the whole scene out and into view
    GLdouble dist = GetCameraDistance(m_pSpace, m_fCubeSize);
    glTranslatef(0.0f, 0.0f, -dist/3);

    //  Do scene rotation
    glMultMatrixf(&m[0][0]);

    float scaledScaleFact = m_fScaleFact / 100.0f;
    glScalef(scaledScaleFact, scaledScaleFact, scaledScaleFact);
  
    if (!IsInOverviewWin()) {
      //  Initialize the names stack
      glInitNames();
      glPushName(0);
    }

    //Tile specific rendering
    RenderTileSpace(&bSelectedCellShown);

  //  Restore the matrix state
	glPopMatrix();	// Modelview matrix

  if (IsInOverviewWin()) {
    DrawOverlay();
  }

  //  Flush and swap buffers so scene is redrawn
  glFlush();
  SwapBuffers();

  if (!IsInOverviewWin()) {
    if (!bSelectedCellShown && m_parentFrame->m_pCellValueWin) {
      m_parentFrame->m_pCellValueWin->SetCurrentCell(m_pMainPanel->m_coordsCurrCell, false, false);
    }
  }
}

float gNewerWidth = 0.0f;
float gNewerHeight = 0.0f;
bool bIn = false;

void TAS3DCanvas::OnSize(wxSizeEvent& event)
{
  // this is also necessary to update the context on some platforms
  wxGLCanvas::OnSize(event);
  
if (!IsInOverviewWin() && !bIn && (gNewerWidth != 0.0f))
{
  bIn = true;
  //  Resize the window and canvas
  this->SetClientSize(gNewerWidth, gNewerHeight);
  m_parent->SetVirtualSize(gNewerWidth, gNewerHeight);

}
bIn = false;

  // set GL viewport (not called by wxGLCanvas::OnSize on all platforms...)
  int w, h;
  GetClientSize(&w, &h);
#ifndef __WXMOTIF__
  if (GetContext())
#endif
  {
    SetCurrent();
    glViewport(0, 0, (GLint) w, (GLint) h);
  }

  // Calculate aspect ratio of the window
	m_fAspect = (GLfloat)w/(GLfloat)h;

  // Set the perspective coordinate system
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  
  // Field of view of 45 degrees, near and far planes 1.0 and 425
  if (m_iViewType == TAS_3D_VIEW_PERSPECTIVE) {
    gluPerspective(m_fFovy, m_fAspect, m_fNear, m_fFar);
  } else {
    glOrtho(-200, 200, -200, 200, -200, 200);
  }
  
  // Modelview matrix reset
//  glMatrixMode(GL_MODELVIEW);
//	glLoadIdentity();

  if (!IsInOverviewWin()) {
    m_parentFrame->m_mgr.Update();
  }
}

void TAS3DCanvas::Reset()
{
  SetScaleFact(100);

  m_fQuat[0] = 0.0f;
  m_fQuat[1] = 0.0f;
  m_fQuat[2] = 0.0f;
  m_fQuat[3] = 1.0f;

  m_parentFrame->RotateAxes(m_fQuat);
  Refresh();
}

void TAS3DCanvas::OnEraseBackground(wxEraseEvent& WXUNUSED(event))
{
  /* Do nothing, to avoid flashing on MSW */
}

// Parse the selection buffer to see which cube was selected
void TAS3DCanvas::ProcessCube(GLuint *pSelectBuff, int hits, int eventType)
{
  GLuint* pCurrRec = pSelectBuff;
  GLuint maxZval = 4294967295;    //  2^32 - 1 (should be the max val)
  int count, i, j, cellInd;

  count = 0;
  
  for (i=0; i<hits; i++) {
    
    if (maxZval > pCurrRec[2]) {
      maxZval = pCurrRec[2];
      
      //  How many names on the name stack
      count = pCurrRec[0];
      
      for (j=0; j<count; j++) {
        //  Bottom of the name stack
        cellInd     = pCurrRec[j+3];
        m_pMainPanel->m_coordsCurrCell = m_pSpace->CoordsFromIndex(cellInd);

        if (eventType == TAS_3D_MOUSE_LDCLICK) {
          m_pMainPanel->m_coordsSelCell = m_pMainPanel->m_coordsCurrCell;
          m_parentFrame->AddCellValueWin();
          m_parentFrame->m_pCellValueWin->SetCurrentCell(m_pMainPanel->m_coordsCurrCell, false, true);

        } else if ((eventType == TAS_3D_MOUSE_LCLICK) && (m_pMainPanel->m_coordsSelCell == m_pMainPanel->m_coordsCurrCell)) {
          m_pMainPanel->m_coordsSelCell = TASCoords();
        } else if (!m_pMainPanel->m_coordsCurrCell.IsValid() &&
                   (m_pMainPanel->m_coordsCurrCell != m_parentFrame->m_pCellValueWin->GetCurrentCell()))
        {
          m_parentFrame->m_pCellValueWin->SetCurrentCell(m_pMainPanel->m_coordsCurrCell, false, false);
        }
      }
    }
    
    //  Move to the next record
    pCurrRec = pCurrRec + 3 + count;
  }
}

// Process the selection, which is triggered by a mouse event
void TAS3DCanvas::ProcessSelection(int xPos, int yPos, int eventType)
{
  if (!m_pSpace) {
    return;
  }

  SetCurrent();

  // Hit counter and viewport storage
  GLint hits, viewport[4];

  int buffSize = (m_pSpace->GetXDim() + m_pSpace->GetYDim() + m_pSpace->GetZDim() + 1) * 4;
  
  // Space for selection buffer
  GLuint* selectBuff = (GLuint*)XtMalloc(sizeof(GLuint) * buffSize + 1);
  selectBuff[0] = 0;
  
  // Setup selection buffer
  glSelectBuffer(buffSize, selectBuff);
  
  // Get the viewport
  glGetIntegerv(GL_VIEWPORT, viewport);
  
  // Switch to projection and save the matrix
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  
    // Change render mode
    glRenderMode(GL_SELECT);
  
    // Establish new clipping volume to be unit cube around
    // mouse cursor point (xPos, yPos) and extending two pixels
    // in the vertical and horzontal direction. Remember OpenGL specifies the
    // y coordinate from the bottom, Windows from the top. So windows position
    // (as measured from the top) subtract the height and you get it in terms 
    // OpenGL Likes.
    glLoadIdentity();
    gluPickMatrix(xPos, viewport[3] - yPos, 1, 1, viewport);
  
    // Apply perspective matrix 
    if (m_iViewType == TAS_3D_VIEW_PERSPECTIVE) {
//MJP: the zFar parameter to gluPerspective seems to need to increase as the
//     size of the tile space increases, otherwise the assembly eventually
//     disappears from sight when it is 'too far back away from the viewer'
      gluPerspective(m_fFovy, m_fAspect, m_fNear, m_fFar);
    } else {
      glOrtho(-200, 200, -200, 200, 1, 200);
    }

    // Draw the scene
    wxPaintEvent pe;
    OnPaint(pe);

    // Collect the hits
    hits = glRenderMode(GL_RENDER);
  
    // Restore the projection matrix
    glMatrixMode(GL_PROJECTION);
  glPopMatrix();

  // If a single hit occured, display the info.
  if(hits > 0) {
    ProcessCube(selectBuff, hits, eventType);
  } else {
    TASCoords emptyCell;
    m_pMainPanel->m_coordsCurrCell = emptyCell;
    if ((m_pMainPanel->m_coordsSelCell == emptyCell) &&
        (emptyCell != m_parentFrame->m_pCellValueWin->GetCurrentCell()))
    {
      m_parentFrame->m_pCellValueWin->SetCurrentCell(emptyCell, false, false);
    }
  }

  XtFree(selectBuff);
}


float fLastScale = 100.0f;

void TAS3DCanvas::DoCanvasResize()
{
  //  Calculate the necessary screen size
  SizerHelper();
  GLdouble newerWidth = m_screenMaxX - m_screenMinX;
  GLdouble newerHeight = m_screenMaxY - m_screenMinY;

  //  Resize the window and canvas
  this->SetClientSize(newerWidth, newerHeight);
  m_parent->SetVirtualSize(newerWidth, newerHeight);
  gNewerWidth = newerWidth;
  gNewerHeight = newerHeight;

  //  Save the last scale factor to be used next time
  fLastScale = m_fScaleFact;
}

float TAS3DCanvas::SetScaleFact(float fScaleFact)
{
  m_fScaleFact = fScaleFact;

  //  Don't allow scaling to become too small
  if (m_fScaleFact < 1.0f) {
    m_fScaleFact = 1.0f;
  }

  //  Don't allow scaling to become too large
  if (m_fScaleFact > 1000.0f) {
    m_fScaleFact = 1000.0f;
  }

  m_parentFrame->SetZoomLevel(m_fScaleFact);

  DoCanvasResize();
  return m_fScaleFact;
}

float TAS3DCanvas::GetScaleFact()
{
  return m_fScaleFact;
}

void TAS3DCanvas::OnMouse(wxMouseEvent& event)
{
  if (event.Dragging()) {
    wxSize sz(GetClientSize());

    if (event.LeftIsDown()) {                         //  Do rotation
      //  Drag in progress, simulate trackball
      float spin_quat[4];
      trackball(spin_quat,
                (2.0 * m_fBeginX - sz.x) / sz.x,
                (sz.y - 2.0 * m_fBeginY) / sz.y,
                (2.0 * event.GetX() - sz.x) / sz.x,
                (sz.y - 2.0 * event.GetY()) / sz.y);
    
      add_quats(spin_quat, m_fQuat, m_fQuat);
    } else if (event.RightIsDown()) {                 //  Do zooming
      float y1 = (sz.y - 2.0 * m_fBeginY) / sz.y;
      float y2 = (sz.y - 2.0 * event.GetY()) / sz.y;
      SetScaleFact(m_fScaleFact + (y2 - y1));
    }

    //  Orientation or zoom factor has changed, redraw mesh
    m_parentFrame->RotateAxes(m_fQuat);
    Refresh();
  } else if (event.LeftDown() /*&& event.ShiftDown()*/) {
    ProcessSelection(event.GetX(), event.GetY(), TAS_3D_MOUSE_LCLICK);
  } else if (event.LeftDClick()) {
    ProcessSelection(event.GetX(), event.GetY(), TAS_3D_MOUSE_LDCLICK);
    Refresh();
  } else if (m_bShowMouseOver) {
    ProcessSelection(event.GetX(), event.GetY(), TAS_3D_MOUSE_OVER);
    Refresh();
  }
  
  event.Skip();

  m_fBeginX = event.GetX();
  m_fBeginY = event.GetY();
  m_parentFrame->UpdateStatusBar();
}

void TAS3DCanvas::OnKeyboard(wxKeyEvent& event)
{
  bool  bDoRot       = false;
  bool  bDoRefresh   = false;
  float phi          = RADS_IN_CIRCLE / 36.0f;
  float axis[3]      = {0.0f, 0.0f, 0.0f };
  float spin_quat[4] = {0.0f, 0.0f, 0.0f, 0.0f };
  
  if (event.m_callbackUserData) {
    wxPoint* pPt = (wxPoint*)event.m_callbackUserData;
    phi = phi * pPt->x;
  }

  if (event.GetKeyCode() == 'Z') {           //  Rotate phi radians around the x-axis (pos or neg)
    if (event.ShiftDown()) {
      axis[2] = 1.0f;
      bDoRot  = true;
    } else {
      axis[2] = -1.0f;
      bDoRot  = true;
    }
  } else if (event.GetKeyCode() == 'Y') {    //  Rotate phi radians around the y-axis (pos or neg)
    if (event.ShiftDown()) {
      axis[1] = 1.0f;
      bDoRot  = true;
    } else {
      axis[1] = -1.0f;
      bDoRot  = true;
    }
  } else if (event.GetKeyCode() == 'X') {    //  Rotate phi radians around the x-axis (pos or neg)
    if (event.ShiftDown()) {
      axis[0] = 1.0f;
      bDoRot  = true;
//MJP: the following is equivalent
//      m_fQuat[0] = m_fQuat[0] + 0.1f;
//      bDoRefresh = true;
    } else {
      axis[0] = -1.0f;
      bDoRot  = true;
    }
  } else if (event.GetKeyCode() == 'R') {    //  Do reset back to original orientation
//MJP: use the fact that this performs a 'reset' to figure out how to rotate around the actual axes above!
    m_fQuat[0] = 0.0f;
    m_fQuat[1] = 0.0f;
    m_fQuat[2] = 0.0f;
    m_fQuat[3] = 1.0f;
    bDoRefresh = true;
  } else if (event.GetKeyCode() == '+') {    //  Zoom in
    SetScaleFact(m_fScaleFact + 25.0f);
    bDoRefresh = true;
  } else if (event.GetKeyCode() == '-') {    //  Zoom out
    SetScaleFact(m_fScaleFact - 25.0f);
    bDoRefresh = true;
  }

//MJP: move the axis to the center of the model?
  if (bDoRot) {
    axis_to_quat(axis, phi, spin_quat);
    add_quats(spin_quat, m_fQuat, m_fQuat);
    bDoRefresh = true;
  }

  if (bDoRefresh) {  
    //  Orientation has changed, redraw mesh
    m_parentFrame->RotateAxes(m_fQuat);
    Refresh();
  }
}

void TAS3DCanvas::SetPerformanceOptions(int perfLevel)
{
  SetCurrent();

  if (perfLevel == TAS_3D_PERF_BESTQUALITY) {
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glEnable(GL_TEXTURE_2D);
    glEnable(GL_ALPHA_TEST);
    glAlphaFunc(GL_GEQUAL, 0.5);

    glEnable(GL_POINT_SMOOTH);
    glEnable(GL_LINE_SMOOTH);
//    glEnable(GL_POLYGON_SMOOTH);

    glEnable(GL_BLEND);
    glEnable(GL_DITHER);

    glShadeModel(GL_SMOOTH);//GL_FLAT

    glHint(GL_PERSPECTIVE_CORRECTION_HINT,  GL_NICEST);
    glHint(GL_POINT_SMOOTH_HINT,            GL_NICEST);
    glHint(GL_LINE_SMOOTH_HINT,             GL_NICEST);
//    glHint(GL_POLYGON_SMOOTH_HINT,          GL_NICEST);

  } else if (perfLevel == TAS_3D_PERF_FASTEST) {
    if (IsInOverviewWin()) {    //  Must always have blending enabled in overview win for transparancy of overlay
      glEnable(GL_BLEND);
    } else {
      glDisable(GL_BLEND);
    }
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_ALPHA_TEST);
    glAlphaFunc(GL_GEQUAL, 0.5);
    
    glDisable(GL_POINT_SMOOTH);
    glDisable(GL_LINE_SMOOTH);
    glDisable(GL_POLYGON_SMOOTH);
    
    glDisable(GL_BLEND);
    glDisable(GL_DITHER);
    
    glShadeModel(GL_FLAT);
    
    glHint(GL_PERSPECTIVE_CORRECTION_HINT,  GL_FASTEST);
    glHint(GL_POINT_SMOOTH_HINT,            GL_FASTEST);
    glHint(GL_LINE_SMOOTH_HINT,             GL_FASTEST);
    glHint(GL_POLYGON_SMOOTH_HINT,          GL_FASTEST);
  }
}

void TAS3DCanvas::InitGL()
{
  SetCurrent();

  m_pTexFont = txfLoadFontFromMem();
//MJP: throw usable error here!
  if (m_pTexFont == NULL) {
    XtError("Failure loading txf data");
    exit(1);
  }

//MJP: lighting sucks!!
  static const GLfloat light0_pos[4]   = { 0.0f, 50.0f, 0.0f, 0.0f };
  static const GLfloat light0_color[4] = { 0.9f, 0.9f, 0.9f, 1.0f };

  static const GLfloat light1_pos[4]   = { 50.0f, 0.0f, 0.0f, 0.0f };
  static const GLfloat light1_color[4] = { 0.9f, 0.9f, 0.9f, 1.0f };

  static const GLfloat light2_pos[4]   = { -50.0f, 0.0f, 0.0f, 0.0f };
  static const GLfloat light2_color[4] = { 0.9f, 0.9f, 0.9f, 1.0f };

  static const GLfloat light3_pos[4]   = { 0.0f, -50.0f, 0.0f, 0.0f };
  static const GLfloat light3_color[4] = { 0.9f, 0.9f, 0.9f, 1.0f };
  
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_MULTISAMPLE);

  //  Remove back faces
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glCullFace(GL_BACK);

  glPolygonMode(GL_FRONT, GL_FILL);
	glFrontFace(GL_CCW);
  
  if (IsInOverviewWin()) {
    SetPerformanceOptions(TAS_3D_PERF_FASTEST);
  } else {
    SetPerformanceOptions(TAS_3D_PERF_BESTQUALITY);
    //SetPerformanceOptions(TAS_3D_PERF_FASTEST);
  }

  dispList = glGenLists(1);

  glEnable(GL_LIGHTING);
  GLfloat global_ambient[] = {0.7, 0.7, 0.7, 1.0};
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);

  //  Light
  glLightfv(GL_LIGHT0, GL_POSITION, light0_pos);
  glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0_color);
  glLightfv(GL_LIGHT1, GL_POSITION, light1_pos);
  glLightfv(GL_LIGHT1, GL_DIFFUSE,  light1_color);
  glLightfv(GL_LIGHT2, GL_POSITION, light2_pos);
  glLightfv(GL_LIGHT2, GL_DIFFUSE,  light2_color);
  glLightfv(GL_LIGHT3, GL_POSITION, light3_pos);
  glLightfv(GL_LIGHT3, GL_DIFFUSE,  light3_color);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glEnable(GL_LIGHT2);
  glEnable(GL_LIGHT3);
  
	// Enable color tracking
  glEnable(GL_COLOR_MATERIAL);
  
  // Set Material properties to follow glColor values
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);

  //  Set line size for wireframes of cubes
  GLfloat sizes[2];
  GLfloat step;
  glGetFloatv(GL_LINE_WIDTH_RANGE, sizes);
  glGetFloatv(GL_LINE_WIDTH_GRANULARITY, &step);
  m_fWireWidth = sizes[0] + (step * 10);
  glLineWidth(m_fWireWidth);
}

