/////////////////////////////////////////////////////////////////////////////
// Name:        TAS3DCanvas.h
// Purpose:     Defines the canvas used for 3-D rendering
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef __TAS3DCanvas__
#define __TAS3DCanvas__

#include "wx/glcanvas.h"
#include "TASTiles.h"
#include "TASPanel.h"

#include <TexFont.h>
#include "trackball.h"

#define TAS_3D_VIEWBY_RANGE       0
#define TAS_3D_VIEWBY_HOOD        1
#define TAS_3D_VIEWBY_UNLIMITED   2

#define TAS_3D_CUBE_NORMAL        0
#define TAS_3D_CUBE_HIGHLIGHTED   1
#define TAS_3D_CUBE_SELECTED      2

#define TAS_3D_VIEW_PERSPECTIVE   0
#define TAS_3D_VIEW_ORTHOGRAPHIC  1

#define TAS_3D_PERF_BESTQUALITY   0
#define TAS_3D_PERF_FASTEST       1

class TASSimulatorFrame;
class TAS3DPanel;

#if wxUSE_GLCANVAS

class TASOGLCanvas : public wxGLCanvas
{
public:
    TASOGLCanvas(wxScrolledWindow* parent, TAS3DPanel* pMainPanel,
                   TASSimulatorFrame* parentFrame, wxWindowID id, const wxPoint& pos,
                   const wxSize& size, long style, const wxString& name);

    virtual void OnKeyboard(wxKeyEvent& event) = 0;

    virtual float SetScaleFact(float fScaleFact) = 0;
    virtual float GetScaleFact() = 0;
    virtual bool IsValid() = 0;
    virtual void Reset() = 0;

    float         m_fQuat[4];           //  Orientation of object
    bool          m_bDoInit;            //  True if initgl not yet called
    bool          m_bShowMouseOver;     //  True if mouseover should be processed
    bool          m_bShowSymbols;       //  True if symbols should be rendered on tile sides
    int           m_iViewType;          //  Either perspective (default) or orthographic
};

class TAS3DCanvas : public TASOGLCanvas
{
public:
    TAS3DCanvas(wxScrolledWindow* parent,
                  TAS3DPanel* pMainPanel,
                  TASSimulatorFrame* parentFrame,
                  wxWindowID id = wxID_ANY,
                  const wxPoint& pos = wxDefaultPosition,
                  const wxSize& size = wxDefaultSize, long style = 0,
                  const wxString& name = wxT("TAS3DCanvas"));    
    ~TAS3DCanvas();

    void OnPaint(wxPaintEvent& event);
    void OnSize(wxSizeEvent& event);
    void OnEraseBackground(wxEraseEvent& event);
    void OnMouse(wxMouseEvent& event);
    void OnKeyboard(wxKeyEvent& event);

    void ProcessSelection(int xPos, int yPos, int eventType);
    void ProcessCube(GLuint *pSelectBuff, int hits, int eventType);
    void DoCanvasResize();
    
    float SetScaleFact(float fScaleFact);
    float GetScaleFact();

    void SetPerformanceOptions(int perfLevel);
    bool IsValid();
    void InitGL();

    //  Functions for rendering cubes of tile space
    bool RenderTileSpace(bool* pbSelectedCellShown);
    void RenderTile(const TASCoords& coords, GLfloat* pxHalf, GLfloat* pyHalf, GLfloat* pzHalf,
                    TASTileType* pTileType, bool* pbSelectedCellShown);

    //  Functions used for rendering font for field symbols on cube faces
    void TextedCube(GLfloat size, char* text, int mode, int rotateCount, wxColour* pColor);
    void WireframeCube(GLfloat size, int mode);
    void CubeSideWithText(GLfloat size, char *text, int len, int mode, int rotateCount, wxColour* pColor);
    void CubeSide(GLfloat size, int mode, bool bWireframe, wxColour* pColor);

    void SetMainPanel(TAS3DPanel* pPanel);
    bool IsInOverviewWin() { return m_bIsOverview; };   //  Returns true if this instance of a TAS3DCanvas is embedded in the overview window
    
    void Reset();

    wxScrolledWindow*   m_parent;         //  Pointer to the window that contains this canvas
    TASSimulatorFrame*  m_parentFrame;    //  Pointer to parent frame
    TexFont*            m_pTexFont;

    TASSpace*           m_pSpace;             //  Pointer to whichever (type of) space is currently being rendered    

    float               m_fBeginX;            //  Position of mouse (x coord)
    float               m_fBeginY;            //  Position of mouse (y coord)
    float               m_fFovy;              //  Field of view in degrees
    float               m_fAspect;            //  Aspect ration (always set to 1.0 for now)
    float               m_fNear;              //  Specifies the	distance from the viewer to the	near
                                              //  clipping plane (always positive)
    float               m_fFar;               //  Specifies the	distance from the viewer to the	far
                                              //  clipping plane (always positive)
    GLfloat             m_fCubeSize;          //  Size of dimensions for cube (tile) sides

    wxColour            m_colWire;            //  The color used to draw the wire edges of cubes
    GLfloat             m_fWireWidth;         //  The width ofthe wireframe edges of cubes

    GLdouble            m_screenMinX;         //  Then minimum X screen coordinate of the rendered space
    GLdouble            m_screenMinY;         //  Then minimum Y screen coordinate of the rendered space
    GLdouble            m_screenMaxX;         //  Then maximum X screen coordinate of the rendered space
    GLdouble            m_screenMaxY;         //  Then maximum Y screen coordinate of the rendered space

protected:
    virtual void DrawOverlay() {};
    void SizerHelper();

    TAS3DPanel*         m_pMainPanel;         //  Pointer to the main 3-D panel
    bool                m_bIsOverview;        //  True if this instance of a TAS3DCanvas is embedded in the overview window
    float               m_fScaleFact;         //  Scaling factor to control zooming

  DECLARE_EVENT_TABLE()
};

#endif // #if wxUSE_GLCANVAS

#endif /* __TAS3DCanvas__ */


