/////////////////////////////////////////////////////////////////////////////
// Name:        TASApp.h
// Purpose:     Declares the main application class
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef TASApp_H
#define TASApp_H

#include "TAS.h"
#include <wx/app.h>
#include <wx/filename.h>
#include <wx/config.h>       // (to let wxWidgets choose a wxConfig class for your platform)
#include <wx/confbase.h>     // (base config class)
#include "TASBreakpoints.h"


#define TAS_2D_MODE             0
#define TAS_3D_MODE             1
#define TAS_TILE_MODE           9
#define TAS_OPT_MODE_0          0
#define TAS_OPT_MODE_1          1
#define TAS_TILECONC_MODE_OFF   0
#define TAS_TILECONC_MODE_ON    1

//  Define TAS_BUILD_3D to include menu option
//  for selecting dimension mode and therefore
//  enable 3-D functionality
#define TAS_BUILD_3D  1


class TASSimulatorFrame;
class TASEditorFrame;

//  Define for user setting value names
#define TASDimModeName            _T("dimMode")
#define TASSimWinName             _T("simWin")
#define TASEditorWinName          _T("editorWin")
#define TASAxesWinName            _T("axesWin")
#define TASUseEditor              _T("useEditor")
#define TASOutputDebugMsgs        _T("debugMsgs")
#define TASReportNondeterm        _T("reportNondeterminism")
#define TASReport1stCondLD        _T("report1stCondLD")
#define TASMaxTileAdditions       _T("maximizeTileAdditions")

#define DefaultTilesetFileName    _T("untitled.tds")
#define DefaultTilespaceFileName  _T("untitled.tdp")

class TASWinSettings
{
public:
  TASWinSettings(wxString strName);
  TASWinSettings(wxString strName, wxPoint ptLoc, wxSize szSize, bool bVisible, bool bMaximized);

  TASWinSettings& operator=(const TASWinSettings& rhs);
  void GetCurrentSettings();
  void SaveCurrentSettings(wxConfig* pConfig);
  void LoadSettings(wxConfig* pConfig);

  wxTopLevelWindow* m_pWin;
  wxString          m_strName;
  wxPoint           m_ptLoc;
  wxSize            m_szSize;
  bool              m_bVisible;
  bool              m_bMaximized;
};

class TASContext : public wxConfig
{
public:
  TASContext();
  ~TASContext() {};

  void LoadSavedSettings();
  void SaveCurrentSettings();

  bool IsAssemblyFileOK();
  bool IsTilesetFileOK();

  //  Window settings
  TASWinSettings    m_setSimWin;
  TASWinSettings    m_setEditorWin;

  //  Modes of operation
  int                 m_iTASDimMode;            //  Number of dimensions (2 or 3)
  int                 m_iTASOptMode;            //  Optimization level
  int                 m_iTASOptModeThreshold;   //  Threshold value (for tileset size) at which optimization level changes
  int                 m_iTASTileConcMode;       //  How to treat tile type concentrations

  //  Input/Output file names
  wxFileName          m_fnTilesetFile;
  wxFileName          m_fnTilespaceFile;

  //  Full wxAui perspective settings for each dockable window
  wxString            m_sCellValuePaneInfo;
  wxString            m_sSpaceConfigPaneInfo;
  wxString            m_sAxesPaneInfo;
  wxString            m_sOverviewPaneInfo;

  //  Full wxAui perspective settings for frames
  wxString            m_sSimFramePersp;
  wxString            m_sEditorFramePersp;

  //  Various UI settings
  wxColour            m_colSpaceBkgrnd;
  wxColour            m_colDefaultTile;
  wxColour            m_colDefaultText;

  unsigned            m_maskSimTileDisplay;       //  Mask to determine which aspects of tiles
                                                  //  are displayed (label, name, glues, strengths)
  unsigned            m_maskEditorTileDisplay;

  //  User preferences
  bool                m_bUseEditor;               //  Whether or not the editor should be created and used
  bool                m_bOutputDebugMsgs;         //  Specifies whether or not debug messages should be output
  bool                m_bReportNondeterminism;    //  This can be toggled on so that potential situations where
                                                  //  tiletype selection is nondeterministic will be reported to the user
  bool                m_bReport1stCondLD;         //  This can be toggled on so that whenever the first condition of
                                                  //  local determinism (tiles bind with strength no greater than tau)
                                                  //  is violated, it will be reported
  bool                m_bMaxTileAdditions;        //  Is this is true, at every simulation step, every tile that can
                                                  //  be added to the current frontier is added
  bool                m_bAsyncDisplay;            //  Allow the simulation to be independent of UI updates.  This allows
                                                  //  the assembly shown in the simulator window to be very out of date,
                                                  //  but drastically speeds up the growth of the assembly

  TASBreakpointSet    m_setBreakpoints;           //  The set of all breakpoints of all types

  //  Specify whether or not the context should be saved
  bool                m_bDoSave;
};

class TASApp : public wxApp
{
public:
  TASApp();
  virtual ~TASApp();

  // OnInit() is called on application startup and is a good place for the app
  // initialization (doing it here and not in the ctor allows to have an error
  // return: if OnInit() returns false, the application terminates)
  virtual bool OnInit();
  bool CreateEditor();
  void ProcessCommandLineFiles(int argc, wxChar* argv[]);
  virtual int OnExit();

  //  Functions to load/reload/reset all aspects of a TAS
  bool LoadTAS(wxString& sTASFileName);                         //  tries to determine tiletypes file from assembly
                                                                //  file and replaces both context values with new,
                                                                //  then updates everything with values from files
  bool LoadTilesetIntoEditor(wxString& sTilesetFileName);       //  loads the tile set into the editor and sets the
                                                                //  filename in the context
  bool LoadTilesetFromSimulator();                              //  loads the tile set from the simulator into the
                                                                //  editor, doesn't change context
  bool LoadTilesetIntoSimulator();                              //  sets the tile set of the simulator equal to the
                                                                //  current tile set of the editor
  bool LoadAssembly(wxString& sAssemblyFileName);               //  loads the tilespace definition from the specified
                                                                //  file and sets the filename in the context

  bool ReloadTAS();                                             //  uses values of filenames stored in the context
                                                                //  to reload space and both tile sets from files
  bool ReloadTilesetIntoEditor();                               //  sets tile space's tile set equal to the current
                                                                //  tile set of the editor
  bool ReloadAssembly();                                        //  uses value of filename from context to reload
                                                                //  the tile space

  bool ResetTAS();                                              //  resets space back to epoch 0 and tile set from
                                                                //  simulator into editor (MJP: not doing the second
                                                                //  part at the moment!)
  bool ResetAssembly();                                         //  resets space back to epoch 0

  bool SaveTASAs(wxString& sTASFileName);                       //  uses string argument as filename base for tile
                                                                //  space file and tile set file and saves the files
                                                                //  and names in the context
  bool SaveTilesetAs(wxString& sTilesetFileName);               //  saves editor's tile sset to specified filename
                                                                //  and name in context
  bool SaveAssemblyAs(wxString& sAssmblyFileName);              //  saves tile space to specified filename and name
                                                                //  in context

  bool SaveTAS();                                               //  use context values to save tile space and editor's
                                                                //  tile set
  bool SaveTileset();                                           //  use context value to save editor's tile set
  bool SaveAssembly();                                          //  use context value to save tile space

  bool NewTAS();                                                //  creates new tile space and tile sets, and blanks
                                                                //  out context values
  bool NewTileset();                                            //  creates new tile set for editor and blanks out
                                                                //  context value
  bool NewAssembly();                                           //  creates new tile space and blanks out context value

  void SetTASDimMode(int mode);
  int  GetTASDimMode();
  void SetDimModeFromTileset();

  void SetTemp(int temp);

  void SaveUserSettings();

  //  Functions for writing text to the output and debug panes of the simulator and editor frames
  void SetSimOutputText(wxString& sMsg, bool bAppend = false, bool bRaiseWin = false);
  void SetSimDebugText(wxString& sMsg, bool bAppend = false, bool bRaiseWin = false);
  void SetEditorOutputText(wxString& sMsg, bool bAppend = false, bool bRaiseWin = false);
  void SetEditorDebugText(wxString& sMsg, bool bAppend = false, bool bRaiseWin = false);


  //  Utility functions used for tracking random number generation and debugging
  static void Srand(unsigned long seed, bool bRandomSeed = false);  //  Wrapper for srand()
//  static int  Rand();                                               //  Wrapper for rand()
  int RandInRange(int n);
  static unsigned long m_randSeed;
  static unsigned long m_randVal;

  TASSimulatorFrame*  m_pSimFrame;
  TASEditorFrame*     m_pEditorFrame;
  bool                m_bShowSplash;     //  Determines is splash screen will be
                                         //  shown - controlled by a command line arg
};

DECLARE_APP(TASApp)

#endif // TASApp_H


