/////////////////////////////////////////////////////////////////////////////
// Name:        TASBreakpoints.cpp
// Purpose:     Defines the classes used for dealing with breakpoints
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------
#include "stdafx.h"
#include "wxprec.h"
#include <wx/tokenzr.h>
#include "TASBreakpoints.h"
#include "TASApp.h"
#include "TASSimFrm.h"


const int TAS_BUTTON_ID_ADD_BREAKPOINT  = 999;
const int TAS_BREAKPOINT_DLG_ID_START   = 1000;

bool TASCoords::SetFromString(wxString& sIn)
{
  Invalidate();
  if (sIn.Contains(_T("("))) sIn = sIn.AfterFirst('(');
  if (sIn.Contains(_T(")"))) sIn = sIn.BeforeLast(')');
  wxStringTokenizer toks(sIn, _T(" ,"), wxTOKEN_STRTOK);
  if (toks.CountTokens() > 0) {
    wxString sTok = toks.GetNextToken();
    long lVal;
    if (sTok.ToLong(&lVal)) {
      x = lVal;
      if (toks.HasMoreTokens()) {
        sTok = toks.GetNextToken();
        if (sTok.ToLong(&lVal)) {
          y = lVal;
          if (toks.HasMoreTokens()) {
            sTok = toks.GetNextToken();
            if (sTok.ToLong(&lVal)) {
              z = lVal;
            }
          } else {
            z = 0;
          }
        }
      }
    }
  }
  if (IsValid()) {
    return true;
  } else {
    return false;
  }
}

wxString TASCoords::ToString()
{
  wxString sRet;
  if (IsValid()) {
    if (g_pContext->m_iTASDimMode == TAS_2D_MODE) {
      sRet = wxString::Format(_T("(%d,%d)"), x, y);
    } else {
      sRet = wxString::Format(_T("(%d,%d,%d)"), x, y, z);
    }
  }
  return sRet;
}

TASBreakpoint::TASBreakpoint(int type)
{
  m_type    = type;
  m_bActive = false;
  m_bIsWild = false;
  m_step    = -1;
  m_coords.Invalidate();
}

wxString TASBreakpoint::GetCondition()
{
  switch (m_type) {
    case Breakpoint_Type_Step:
      if (m_step == -1) {
        return wxString::Format(_T(""));
      } else {
        return wxString::Format(_T("%d%s"), m_step, (m_bIsWild ? _T("*") : _T("")));
      }
    case Breakpoint_Type_Location:
      return m_coords.ToString();
    case Breakpoint_Type_Tiletype:
      return m_sTiletypeName;
    default:
      return wxString(_T(""));
  }
}

TASBreakpointUI::TASBreakpointUI(int type)
{
  InvalidateUI();
}

TASBreakpointUI::TASBreakpointUI(const TASBreakpoint& bp)
{
  m_type          = bp.m_type;
  m_bActive       = bp.m_bActive;
  m_bIsWild       = bp.m_bIsWild;
  m_step          = bp.m_step;
  m_coords        = bp.m_coords;
  m_sTiletypeName = bp.m_sTiletypeName;

  InvalidateUI();
}

void TASBreakpointUI::InvalidateUI()
{
  m_num               = -1;
  m_pSizer            = NULL;
  m_pTextNum          = NULL;
  m_pCheckActive      = NULL;
  m_pChoiceType       = NULL;
  m_pTextCondition    = NULL;
  m_pButtonDelete     = NULL;
}

void TASBreakpointUI::DeleteUI()
{
  if (m_pTextNum) delete m_pTextNum;
  if (m_pCheckActive) delete m_pCheckActive;
  if (m_pChoiceType) delete m_pChoiceType;
  if (m_pTextCondition) delete m_pTextCondition;
  if (m_pButtonDelete) delete m_pButtonDelete;
  if (m_pSizer) delete m_pSizer;

  InvalidateUI();
}

void TASBreakpointSet::ProcessSet()
{
  m_vectBreakpointSteps.clear();
  m_vectBreakpointWildSteps.clear();
  m_vectBreakpointLocations.clear();
  m_vectBreakpointTiletypes.clear();

  for (size_t i=0; i<size(); i++) {
    if (at(i).m_bActive) {
      if (at(i).m_type == Breakpoint_Type_Step) {
        if (at(i).m_bIsWild) {
          m_vectBreakpointWildSteps.push_back(at(i).m_step);
        } else {
          m_vectBreakpointSteps.push_back(at(i).m_step);
        }
      } else if (at(i).m_type == Breakpoint_Type_Location) {
        m_vectBreakpointLocations.push_back(at(i).m_coords);
      } else if (at(i).m_type == Breakpoint_Type_Tiletype) {
        m_vectBreakpointTiletypes.push_back(at(i).m_sTiletypeName);
      }
    }
  }
}

TASBreakpointUI* TASBreakpointUIList::FindMatchingEntry(wxWindow* pWin)
{
  for (size_t i=0; i<size(); i++) {
    if ((at(i).m_pSizer         == (wxBoxSizer*)pWin) ||
        (at(i).m_pTextNum       == (wxStaticText*)pWin) ||
        (at(i).m_pCheckActive   == (wxCheckBox*)pWin) ||
        (at(i).m_pChoiceType    == (wxChoice*)pWin) ||
        (at(i).m_pTextCondition == (wxTextCtrl*)pWin) ||
        (at(i).m_pButtonDelete  == (wxButton*)pWin))
    {
      return &at(i);
    }
  }
  return NULL;
}


IMPLEMENT_CLASS(TASBreakpointWindow, wxDialog)
BEGIN_EVENT_TABLE(TASBreakpointWindow, wxDialog)
  EVT_BUTTON(wxID_OK, TASBreakpointWindow::OnOk)
  EVT_BUTTON(wxID_ANY, TASBreakpointWindow::OnButton)
END_EVENT_TABLE()

TASBreakpointWindow::TASBreakpointWindow(wxWindow* pParent, TASBreakpointSet* setBPs)
                      : wxDialog(pParent, wxID_ANY, _T(""), wxDefaultPosition)
{
  if (setBPs) {
    for (size_t i=0; i<setBPs->size(); i++) {
      TASBreakpointUI bpui(setBPs->at(i));
      m_listBreakpoints.push_back(bpui);
    }
  }

  wxString sTitle = _T("Manage Breakpoints");
  SetTitle(sTitle);

  m_pMainSizer = new wxBoxSizer(wxVERTICAL);

  m_pMainSizer->AddSpacer(10);

  wxBoxSizer* pLabels = new wxBoxSizer(wxHORIZONTAL);
  wxStaticText* pText = new wxStaticText(this, wxID_ANY, _T("     Active         Type                  Condition"));
  pLabels->AddSpacer(5);
  pLabels->Add(pText, 0, wxTOP, 5);

  m_pMainSizer->Add(pLabels);

  m_pListSizer = NULL;
  BuildBreakpointListUI();

  m_pMainSizer->AddSpacer(3);
  wxBoxSizer* pAddSizer = new wxBoxSizer(wxHORIZONTAL);
  wxButton* pAdd = new wxButton(this, TAS_BUTTON_ID_ADD_BREAKPOINT, _T("Add"), wxDefaultPosition, wxSize(50, 20));
  pAddSizer->Add(pAdd, 0, wxLEFT, 226);
  m_pMainSizer->Add(pAddSizer);

  m_pMainSizer->AddSpacer(10);
  wxBoxSizer* pButtonSizer = new wxBoxSizer(wxHORIZONTAL);
  wxButton* pOK = new wxButton(this, wxID_OK, _T("OK"));
  pButtonSizer->Add(pOK, 0, wxLEFT, 140);
  wxButton* pCancel = new wxButton(this, wxID_CANCEL, _T("Cancel"));
  pButtonSizer->Add(pCancel);

  m_pMainSizer->Add(pButtonSizer);
  m_pMainSizer->AddSpacer(5);

  SetAutoLayout(TRUE);
  SetSizer(m_pMainSizer);

  m_pMainSizer->SetSizeHints(this);
  m_pMainSizer->Fit(this);
}

void TASBreakpointWindow::OnButton(wxCommandEvent& event)
{
  TASBreakpointUI* pUI = m_listBreakpoints.FindMatchingEntry((wxWindow*)event.GetEventObject());

  if (event.GetId() == TAS_BUTTON_ID_ADD_BREAKPOINT) {
    TASBreakpointUI bp;
    bp.m_bActive = true;
    AddBreakpoint(bp);

  } else if (event.GetId() == wxID_CANCEL) {
    EndModal(wxID_CANCEL);
  } else if (pUI && (pUI->m_pButtonDelete == (wxButton*)event.GetEventObject())) {
    int doDelete = wxMessageBox(_T("Are you sure you want to delete this breakpoint?"), 
                                _T("Confirm delete breakpoint"), 
                                wxYES_NO, this);

    if (doDelete == wxYES) {

      //  Remove the breakpoint from the list
      TASBreakpointUIList listBreakpoints;
      for (size_t i=0; i<m_listBreakpoints.size(); i++) {
        if (i != pUI->m_num) {
          m_listBreakpoints[i].m_num = listBreakpoints.size();
          m_listBreakpoints[i].m_pTextNum->SetLabel(wxString::Format(_T("%d"), (listBreakpoints.size() + 1)));
          listBreakpoints.push_back(m_listBreakpoints[i]);
        } else {
          m_pListSizer->Detach(m_listBreakpoints[i].m_pSizer);
          m_listBreakpoints[i].DeleteUI();
        }
      }
      m_listBreakpoints.clear();
      m_listBreakpoints = listBreakpoints;

      //  Re-layout the window
      Fit();
    }
  }
}

void TASBreakpointWindow::BuildBreakpointListUI()
{
  //  Delete all UI associated with the list
  if (m_pListSizer) {
    delete m_pListSizer;
  }
  m_pListSizer = new wxBoxSizer(wxVERTICAL);

  TASBreakpointUIList listBPs = m_listBreakpoints;
  m_listBreakpoints.clear();
  for (size_t i=0; i<listBPs.size(); i++) {
    listBPs[i].InvalidateUI();
    AddBreakpoint(listBPs[i]);
  }
  m_pMainSizer->Add(m_pListSizer);
}

void TASBreakpointWindow::OnOk(wxCommandEvent& WXUNUSED(event))
{
  wxString sWarning;

  for (size_t i=0; i<m_listBreakpoints.size(); i++) {
    bool bActive = m_listBreakpoints[i].m_pCheckActive->IsChecked();
    wxString sCondition = m_listBreakpoints[i].m_pTextCondition->GetValue();
    wxChoice* pType = m_listBreakpoints[i].m_pChoiceType;
    m_listBreakpoints[i].m_type = pType->GetSelection();
    if (m_listBreakpoints[i].m_type == Breakpoint_Type_Step) {
      m_listBreakpoints[i].m_bActive = bActive;
      if (sCondition.Contains(_T("*"))) {
        m_listBreakpoints[i].m_bIsWild = true;
        sCondition.Replace(_T("*"), _T(""));
      }
      long lVal;
      if (sCondition.ToLong(&lVal)) {
        m_listBreakpoints[i].m_step = (int)lVal;
      } else if (m_listBreakpoints[i].m_bActive) {
        m_listBreakpoints[i].m_bActive = false;
        m_listBreakpoints[i].m_pCheckActive->SetValue(false);
        sWarning += wxString::Format(_T("Number: '%d', Type: 'Step', Condition: '%s'\n"), (i + 1), m_listBreakpoints[i].m_pTextCondition->GetValue().c_str());
      }
    } else if (m_listBreakpoints[i].m_type == Breakpoint_Type_Location) {
      m_listBreakpoints[i].m_bActive = bActive;
      m_listBreakpoints[i].m_coords.SetFromString(sCondition);
      if (!m_listBreakpoints[i].m_coords.IsValid() && m_listBreakpoints[i].m_bActive) {
        m_listBreakpoints[i].m_bActive = false;
        m_listBreakpoints[i].m_pCheckActive->SetValue(false);
        sWarning += wxString::Format(_T("Number: '%d', Type: 'Location', Condition: '%s'\n"), (i + 1), m_listBreakpoints[i].m_pTextCondition->GetValue().c_str());
      }
    } else if (m_listBreakpoints[i].m_type == Breakpoint_Type_Tiletype) {
      m_listBreakpoints[i].m_bActive = bActive;
      m_listBreakpoints[i].m_sTiletypeName = sCondition;
      if ((m_listBreakpoints[i].m_sTiletypeName.Len() <= 0) && m_listBreakpoints[i].m_bActive) {
        m_listBreakpoints[i].m_bActive = false;
        m_listBreakpoints[i].m_pCheckActive->SetValue(false);
        sWarning += wxString::Format(_T("Number: '%d', Type: 'Tile type', Condition: '%s'\n"), (i + 1), m_listBreakpoints[i].m_pTextCondition->GetValue().c_str());
      }
    }
  }

  bool bExit = true;
  if (sWarning.Len() > 0) {
    sWarning = wxString(_T("Warning:  The following breakpoints are not\nvalid and have been disabled:\n\n")) + sWarning;
    g_pTASApp->SetSimOutputText(sWarning, true);
    int answer = wxMessageBox(sWarning, _T("Invalid breakpoints"), wxICON_WARNING | wxOK | wxCANCEL);
    if (answer == wxCANCEL) {
      bExit = false;
    }
  }

  if (bExit) {
    EndModal(wxID_OK);
  }
}

void TASBreakpointWindow::AddBreakpoint(TASBreakpointUI& bp)
{
  bp.m_num = m_listBreakpoints.size();

  wxString arrTypes[] = {_T("Step"), _T("Location"), _T("Tile type") };
  
  bp.m_pSizer = new wxBoxSizer(wxHORIZONTAL);
  bp.m_pSizer->AddSpacer(10);

  bp.m_pTextNum = new wxStaticText(this, wxID_ANY, wxString::Format(_T("%d"), (bp.m_num + 1)));
  bp.m_pSizer->Add(bp.m_pTextNum, 0, wxTOP, 5);
  bp.m_pSizer->AddSpacer(12);

  bp.m_pCheckActive = new wxCheckBox(this, wxID_ANY, _T(""));
  bp.m_pCheckActive->SetValue(bp.m_bActive);
  bp.m_pSizer->Add(bp.m_pCheckActive, 0, wxTOP, 5);
  bp.m_pSizer->AddSpacer(15);

  bp.m_pChoiceType = new wxChoice(this, wxID_ANY, wxDefaultPosition, wxSize(80, 20), 3, arrTypes);
  bp.m_pChoiceType->SetSelection((bp.GetType() != Breakpoint_Type_None) ? bp.GetType() : Breakpoint_Type_Step);
  bp.m_pSizer->Add(bp.m_pChoiceType);

  bp.m_pTextCondition = new wxTextCtrl(this, wxID_ANY, _T(""), wxDefaultPosition, wxSize(80, 20));
  bp.m_pTextCondition->SetValue(bp.GetCondition());
  bp.m_pSizer->Add(bp.m_pTextCondition, 0, wxLEFT, 5);

  bp.m_pButtonDelete = new wxButton(this, wxID_ANY, _T("Delete"), wxDefaultPosition, wxSize(50, 20));
  bp.m_pSizer->Add(bp.m_pButtonDelete, 0, wxLEFT, 5);

  m_pListSizer->Add(bp.m_pSizer, 0, wxTOP, 3);

  m_pMainSizer->Fit(this);

  m_listBreakpoints.push_back(bp);
}

