/////////////////////////////////////////////////////////////////////////////
// Name:        TASTiles.cpp
// Purpose:     Defines classes for tiles, tile space, etc.
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////


// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------
#include "stdafx.h"

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

// for all others, include the necessary headers (this file is usually all you
// need because it includes almost all "standard" wxWidgets headers)
#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include <algorithm>
#include <map>
#include <wx/tokenzr.h>
#include "TAS.h"
#include "TASApp.h"
#include "TASFrmBase.h"
#include "TASSimFrm.h"
#include "TASTiles.h"
#include <wx/hashmap.h>
#include "TASDebug.h"


//  Create the global tile space pointer
TASTileSpace*     g_pTileSpace      = NULL;
TASTileTypeSetEx* g_pEditorTileset  = NULL;

TASCoords coordsNbrhd[7]    = { TASCoords(0, 0, 0), TASCoords(1, 0, 0), TASCoords(0, 1, 0), TASCoords(-1, 0, 0), TASCoords(0, -1, 0), TASCoords(0, 0, 1), TASCoords(0, 0, -1) };
int TASOppositeDirection[]  = { TAS_SOUTH, TAS_WEST, TAS_NORTH, TAS_EAST, TAS_DOWN, TAS_UP };

TASTileType g_EmptyTileType;
int g_tileSize = 100;

bool TASIsTileType(size_t t)
{
  return (t < TAS_TILE_FRONTIERDEAD);
}

wxString TASDirectionNumToString(int dirNum, bool bAllCaps)
{
  wxString retStr = _T("Invalid Direction");

  switch (dirNum) {
    case TAS_NORTH:
      if (bAllCaps) {
        retStr = _T("NORTH");
      } else {
        retStr = _T("North");
      }
      break;
    case TAS_SOUTH:
      if (bAllCaps) {
        retStr = _T("SOUTH");
      } else {
        retStr = _T("South");
      }
      break;
    case TAS_EAST:
      if (bAllCaps) {
        retStr = _T("EAST");
      } else {
        retStr = _T("East");
      }
      break;
    case TAS_WEST:
      if (bAllCaps) {
        retStr = _T("WEST");
      } else {
        retStr = _T("West");
      }
      break;
    case TAS_UP:
      if (bAllCaps) {
        retStr = _T("UP");
      } else {
        retStr = _T("Up");
      }
      break;
    case TAS_DOWN:
      if (bAllCaps) {
        retStr = _T("DOWN");
      } else {
        retStr = _T("Down");
      }
      break;
    default:
      break;
  }
  return retStr;
}

#include <wx/arrimpl.cpp>
WX_DEFINE_OBJARRAY(TASLocationTransitionList);

TASTileEpoch::TASTileEpoch()
{
  ConstructorHelper();
}

TASTileEpoch::TASTileEpoch(int temp)
{
  ConstructorHelper();
  m_temp = temp;
}

void TASTileEpoch::ConstructorHelper()
{
  m_arrBases[TAS_BEGIN_TILES]         =  0;
  m_arrBases[TAS_BEGIN_DEADFRONTIER]  =  0;
  m_arrBases[TAS_BEGIN_FRONTIER]      =  0;
  m_arrBases[TAS_BEGIN_EMPTY]         =  0;
  m_temp                              = -1;
}

size_t TASTileEpoch::GetEnd() {
  return m_listTransitions.GetCount();
}

bool TASTileEpoch::AddTransition(TASLocationTransition& trans)
{
  //  See if there is already a transition for the given location
  for (size_t i=0; i<m_listTransitions.GetCount(); i++) {
    if (m_listTransitions[i].m_coords == trans.m_coords) {
      //  If the existing transition results in the same type simply
      //  replace it
      if ((TASIsTileType(m_listTransitions[i].m_typeEnd) && TASIsTileType(trans.m_typeEnd)) ||
          (m_listTransitions[i].m_typeEnd == trans.m_typeEnd))
      {
        m_listTransitions[i] = trans;
        return true;
      } else {  //  If it results in something else, delete the existing
                //  transition and below we'll create a new one
        RemoveTransition(trans);
      }
    }
  }

  //  Create a new transition to the necessary type
  if (trans.m_typeEnd == TAS_TILE_NOT_FOUND) {
    m_listTransitions.Add(trans);
    return true;
  } else if (trans.m_typeEnd == TAS_TILE_FRONTIER) {
    m_listTransitions.Insert(trans, m_arrBases[TAS_BEGIN_EMPTY]);
    m_arrBases[TAS_BEGIN_EMPTY]++;
    return true;
  } else if (trans.m_typeEnd == TAS_TILE_FRONTIERDEAD) {
    m_listTransitions.Insert(trans, m_arrBases[TAS_BEGIN_FRONTIER]);
    m_arrBases[TAS_BEGIN_FRONTIER]++;
    m_arrBases[TAS_BEGIN_EMPTY]++;
    return true;
  } else if (TASIsTileType(trans.m_typeEnd)) {
    m_listTransitions.Insert(trans, m_arrBases[TAS_BEGIN_DEADFRONTIER]);
    m_arrBases[TAS_BEGIN_DEADFRONTIER]++;
    m_arrBases[TAS_BEGIN_FRONTIER]++;
    m_arrBases[TAS_BEGIN_EMPTY]++;
    return true;
  }

  return false;
}

bool TASTileEpoch::RemoveTransition(TASLocationTransition& trans)
{
  int index = m_listTransitions.Index(trans);

  if (index == wxNOT_FOUND) {
    return false;
  } else {
    m_listTransitions.RemoveAt(index);
    if (trans.m_typeEnd == TAS_TILE_FRONTIER) {
      m_arrBases[TAS_BEGIN_EMPTY]--;
    } else if (trans.m_typeEnd == TAS_TILE_FRONTIERDEAD) {
      m_arrBases[TAS_BEGIN_FRONTIER]--;
      m_arrBases[TAS_BEGIN_EMPTY]--;
    } else if (TASIsTileType(trans.m_typeEnd)) {
      m_arrBases[TAS_BEGIN_DEADFRONTIER]--;
      m_arrBases[TAS_BEGIN_FRONTIER]--;
      m_arrBases[TAS_BEGIN_EMPTY]--;
    }
    return true;
  }
}

/*
bool TASTileEpoch::RemoveTransitionsToType(int baseType)
{
  int rangeEnd = 0;
  if (baseType == TAS_BEGIN_EMPTY) {
    rangeEnd = m_listTransitions.GetCount();
  } else {
    rangeEnd = m_arrBases[baseType+1];
  }

  int num = rangeEnd - m_arrBases[baseType];
  for (size_t i=m_arrBases[baseType]; i<rangeEnd; i++) {
MJP: this is incorrect!!    m_listTransitions.RemoveAt(i);
  }

  for (int j=TAS_BEGIN_TILES; j<=TAS_BEGIN_EMPTY; j++) {
    if (baseType < j) {
      m_arrBases[j] -= num;
    }
  }
}*/

bool TASTileEpoch::RemoveTransitionAtLocation(TASCoords& coords)
{
  for (size_t i=0; i<m_listTransitions.GetCount(); i++) {
    if (m_listTransitions[i].m_coords == coords) {
      TASLocationTransition trans(m_listTransitions[i]);
      return RemoveTransition(trans);
    }
  }
  return false;
}

void TASTileEpoch::ClearTransitions()
{
  m_listTransitions.Empty();
}

bool TASTileSide::operator==(const TASTileSide& right)
{
  if (m_hashNum == right.m_hashNum) {
    return true;
  }
  return false;
}

TASTileSideHash::TASTileSideHash(TASTileSide& tsNorth, TASTileSide& tsEast,
                                 TASTileSide& tsSouth, TASTileSide& tsWest,
                                 TASTileSide& tsUp,    TASTileSide& tsDown)
{
  m_arr[TAS_NORTH] = tsNorth;
  m_arr[TAS_EAST]  = tsEast;
  m_arr[TAS_SOUTH] = tsSouth;
  m_arr[TAS_WEST]  = tsWest;
  m_arr[TAS_UP]    = tsUp;
  m_arr[TAS_DOWN]  = tsDown;

  CreateHash();
}

int orderOptDirs[6] = { TAS_SOUTH, TAS_WEST, TAS_NORTH, TAS_EAST, TAS_DOWN, TAS_UP };
void TASTileSideHash::CreateHash()
{
  m_strHash = _T("");
  for (int i=TAS_NORTH; i<=TAS_DOWN; i++) {
    int d = orderOptDirs[i];
    m_strHash += wxString::Format(_T("%d:"), m_arr[d].m_iStrength);
    if (m_arr[d].m_hashNum == TAS_TILE_NOT_FOUND) {
      m_strHash += _T("E");
    } else if (m_arr[d].m_hashNum == TAS_TILE_FRONTIER) {
      m_strHash += _T("F");
    } else if (m_arr[d].m_hashNum == TAS_TILE_FRONTIERDEAD) {
      m_strHash += _T("D");
    } else {
      m_strHash += wxString::Format(_T("%d"), m_arr[d].m_hashNum);
    }
    if (i != TAS_DOWN) {
      m_strHash += _T(";");
    }
  }
}

bool TASTileSideHash::operator<(const TASTileSideHash& tshRight) const {
  return (m_strHash.Cmp(tshRight.m_strHash) < 0);
}

bool TASTileSideHash::operator==(const TASTileSideHash& tshRight) {
  return (m_strHash.Cmp(tshRight.m_strHash) == 0);
}

TASTileType::TASTileType(int dims)
{
  m_dims          = dims;
  m_concentration = 1;
  m_colTile       = wxNullColour;
  m_colText       = wxNullColour;
}

TASTileType::TASTileType(const TASTileType& tt)
{
  SetEqual(tt);
}

void TASTileType::SetEqual(const TASTileType& tt)
{
  m_dims          = tt.m_dims;
  m_concentration = tt.m_concentration;
  m_sName         = tt.m_sName;
  m_sLabel        = tt.m_sLabel;
  m_sComment      = tt.m_sComment;
  for (int i=TAS_NORTH; i<=TAS_DOWN; i++) {
    m_sides[i]    = tt.m_sides[i];
  }
  m_image         = tt.m_image;
  m_colTile       = tt.m_colTile;
  m_colText       = tt.m_colText;
  m_sImagePath    = tt.m_sImagePath;

  InvalidateBitmaps();
}

TASTileType::TASTileType(const wxString& sTTDefn, size_t* index)
{
  m_dims          = 2;
  m_concentration = 1;
  m_colTile       = wxNullColour;
  m_colText       = wxNullColour;

  size_t len      = sTTDefn.Len();
  size_t newIndex = *index;
  while ((newIndex < len) && (sTTDefn[newIndex] != '\n')) {
    newIndex++;
  }
  wxString sLine = sTTDefn.SubString(*index, newIndex);
  newIndex++;
  *index = newIndex;
  while (sLine.length() > 0) {
    sLine.Trim().Trim(false);

    //  Parse current line and get value
    if (sLine.StartsWith(_T("TILENAME"))) {
      m_sName = sLine.SubString(8, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("LABEL"))) {
      m_sLabel = sLine.SubString(5, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("NORTHBIND"))) {
      sLine = sLine.SubString(9, sLine.Len()).Trim(false);
      long strength = 0;
      if (sLine.IsNumber() && sLine.ToLong(&strength)) {
        m_sides[TAS_NORTH].m_iStrength = strength;
      }
    } else if (sLine.StartsWith(_T("EASTBIND"))) {
      sLine = sLine.SubString(8, sLine.Len()).Trim(false);
      long strength = 0;
      if (sLine.IsNumber() && sLine.ToLong(&strength)) {
        m_sides[TAS_EAST].m_iStrength = strength;
      }
    } else if (sLine.StartsWith(_T("SOUTHBIND"))) {
      sLine = sLine.SubString(9, sLine.Len()).Trim(false);
      long strength = 0;
      if (sLine.IsNumber() && sLine.ToLong(&strength)) {
        m_sides[TAS_SOUTH].m_iStrength = strength;
      }
    } else if (sLine.StartsWith(_T("WESTBIND"))) {
      sLine = sLine.SubString(8, sLine.Len()).Trim(false);
      long strength = 0;
      if (sLine.IsNumber() && sLine.ToLong(&strength)) {
        m_sides[TAS_WEST].m_iStrength = strength;
      }
    } else if (sLine.StartsWith(_T("UPBIND"))) {
      m_dims = 3;
      sLine = sLine.SubString(6, sLine.Len()).Trim(false);
      long strength = 0;
      if (sLine.IsNumber() && sLine.ToLong(&strength)) {
        m_sides[TAS_UP].m_iStrength = strength;
      }
    } else if (sLine.StartsWith(_T("DOWNBIND"))) {
      m_dims = 3;
      sLine = sLine.SubString(8, sLine.Len()).Trim(false);
      long strength = 0;
      if (sLine.IsNumber() && sLine.ToLong(&strength)) {
        m_sides[TAS_DOWN].m_iStrength = strength;
      }
    } else if (sLine.StartsWith(_T("NORTHLABEL"))) {
      m_sides[TAS_NORTH].m_sLabel = sLine.SubString(10, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("EASTLABEL"))) {
      m_sides[TAS_EAST].m_sLabel = sLine.SubString(9, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("SOUTHLABEL"))) {
      m_sides[TAS_SOUTH].m_sLabel = sLine.SubString(10, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("WESTLABEL"))) {
      m_sides[TAS_WEST].m_sLabel = sLine.SubString(9, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("UPLABEL"))) {
      m_dims = 3;
      m_sides[TAS_UP].m_sLabel = sLine.SubString(7, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("DOWNLABEL"))) {
      m_dims = 3;
      m_sides[TAS_DOWN].m_sLabel = sLine.SubString(9, sLine.Len()).Trim(false);
    } else if (sLine.StartsWith(_T("CONCENTRATION"))) {
      sLine = sLine.SubString(13, sLine.Len()).Trim(false);
      long concentration = 0;
      if (sLine.IsNumber() && sLine.ToLong(&concentration)) {
       m_concentration = concentration;
      }
    } else if (sLine.StartsWith(_T("TILECOLOR"))) {
      wxString sColor = sLine.SubString(9, sLine.Len()).Trim(false);
      if (sColor.Len() > 0) {
        if (sColor.StartsWith(_T("RGB("))) {
          m_colTile.Set(sColor);
        } else {
          m_colTile = sColor;
        }
      }
    } else if (sLine.StartsWith(_T("TEXTCOLOR"))) {
      wxString sColor = sLine.SubString(9, sLine.Len()).Trim(false);
      if (sColor.Len() > 0) {
        if (sColor.StartsWith(_T("RGB("))) {
          m_colText.Set(sColor);
        } else {
          m_colText = sColor;
        }
      }
    } else if (sLine.StartsWith(_T("icon"))) {
      wxFileName fnImg;
      fnImg.AssignDir(g_pContext->m_fnTilesetFile.GetPath());
      m_sImagePath = sLine.SubString(4, sLine.Len()).Trim(false);
      fnImg.SetFullName(m_sImagePath);
      m_image = wxImage(fnImg.GetFullPath());
      if (!m_image.IsOk()) {
        wxString sBadImage = wxString::Format(_T("Failed to load image for tiletype '%s'\n"), m_sName.c_str());
        g_pTASApp->SetEditorOutputText(sBadImage, true, true);
      }
    } else if (sLine.StartsWith(_T("CREATE"))) {
      break;
    }

    while ((newIndex < len) && (sTTDefn[newIndex] != '\n')) {
      newIndex++;
    }
    sLine = sTTDefn.SubString(*index, newIndex);
    newIndex++;
    *index = newIndex;
  }
}

TASTileType::~TASTileType()
{
}

bool TASTileType::IsValid()
{
return (m_sName.Len() > 0);
/*
  if (m_dims == 2) {
    if ((m_iStrengths[4] != -99) || (m_iStrengths[5] != -99) ||
      (m_glues[4].m_strLabel.Len() > 0 || m_glues[5].m_strLabel.Len() >0))
    {
      return false;
    } else {
      return true;
    }
  } else if (m_dims == 3) {
    return true;
  }
  return false;
*/
}

bool TASTileType::IsEqual(TASTileType* pTtype, bool bCompareNameAndLabel, bool bAll)
{
  if (!pTtype || (pTtype->m_dims != m_dims)) {
    return false;
  }

  if (bCompareNameAndLabel || bAll) {
    if ((pTtype->m_sName.CompareTo(m_sName.c_str()) != 0) ||
        (pTtype->m_sLabel.CompareTo(m_sLabel.c_str()) != 0))
    {
      return false;
    }
  }

  if (bAll) {
    if ((pTtype->m_colText != m_colText) || (pTtype->m_colTile != m_colTile))
    {
      return false;
    }
  }

  for (int i=TAS_NORTH; i<TAS_DOWN; i++) {
    if (m_sides[i] != pTtype->m_sides[i]) {
      return false;
    }
  }

  return true;
}

//  Rotation is done in the clockwise direction
bool TASTileType::Rotate(int degrees)
{
  if (degrees != 0 && degrees != 90 && degrees != 180 && degrees != 270 && degrees != 360)
  {
    return false;
  }
  
  TASTileSide ts[4];
  for (int j=0; j<4; j++) {
    ts[j] = m_sides[j];
  }

  int rotCount = degrees / 90;
  for (int i=0; i<4; i++) {
    m_sides[(i+rotCount)%4] = ts[i];
  }
  InvalidateBitmaps();
  return true;
}

wxString TASTileType::Serialize()
{
  wxString outStr;
  
  outStr += _T("TILENAME ") + m_sName + _T("\n");
  outStr += _T("LABEL ") + m_sLabel + _T("\n");
  if (m_concentration > 1) {
    outStr += wxString::Format(_T("CONCENTRATION %d\n"), m_concentration);
  }
  for (int i=TAS_NORTH; i<=(m_dims==2?TAS_WEST:TAS_DOWN); i++) {
    outStr += TASDirectionNumToString(i, true) + _T("BIND ") + wxString::Format(_T("%d"), m_sides[i].m_iStrength) + _T("\n");
  }
  for (int i=TAS_NORTH; i<=(m_dims==2?TAS_WEST:TAS_DOWN); i++) {
    outStr += TASDirectionNumToString(i, true) + _T("LABEL ") + m_sides[i].m_sLabel + _T("\n");
  }
  if (m_colTile != wxNullColour) {
    outStr += _T("TILECOLOR ") + m_colTile.GetAsString() + _T("\n");
  }
  if (m_colText != wxNullColour) {
    outStr += _T("TEXTCOLOR ") + m_colText.GetAsString() + _T("\n");
  }
  if (m_image.IsOk() && (m_sImagePath.Length() > 0)) {
    outStr += _T("icon ") + m_sImagePath + _T("\n");
  }
  outStr += _T("CREATE\n\n");

  return outStr;
}

wxBitmap& TASTileType::GetBitmap(int dim)
{
  if (m_vecBitmaps.size() <= dim) {
    m_vecBitmaps.resize(dim+1);
  }
  return m_vecBitmaps[dim].m_bitmap;
}

void TASTileType::InvalidateBitmaps()
{
  m_vecBitmaps.clear();
  m_bitmapFlat3D = wxBitmap();
}

wxFont fontLabel(12, wxSWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD, false);
wxFont fontLabelSVG(6, wxSWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD, false);

wxFont fontGlues(8, wxSWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL, false);
wxFont fontGluesSVG(4, wxSWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL, false);
wxFont fontGlues3D(12, wxSWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL, false);

bool bHighQuality = true;  //  I think this should stay set to 'true', but performance needs to be tested more extensively

//                  TAS_NORTH, TAS_EAST, TAS_SOUTH, TAS_WEST, TAS_UP, TAS_DOWN
int xDirOffset[] = {        1,        2,         1,        0,      1,        1};
int yDirOffset[] = {        1,        1,         3,        1,      0,        2};

void TASTileType::DrawFlat3DSide(wxDC* pDC, wxPoint& pt, int dim, int side)
{
  int strength = m_sides[side].m_iStrength;
  if (strength > 0) {
    m_pen.SetStyle(wxSOLID);
    pDC->SetPen(m_pen);
  } else {
    m_pen.SetStyle(wxDOT);
    pDC->SetPen(m_pen);
  }

  int xOffset = pt.x + (dim * xDirOffset[side]);
  int yOffset = pt.y + (dim * yDirOffset[side]);
  pDC->DrawRectangle(xOffset, yOffset, dim, dim);

  if (strength > 1) {
    int dblBndSep = 4;
    pDC->DrawRectangle(xOffset + dblBndSep,
                       yOffset + dblBndSep,
                       dim - (2*dblBndSep),
                       dim - (2*dblBndSep));
  }

  if (strength > 0) {
    wxCoord w, h;
    pDC->GetTextExtent(m_sides[side].m_sLabel, &w, &h);

    wxCoord xStart = xOffset + (dim / 2) - (w / 2);
    wxCoord yStart = yOffset + (dim / 2) - (h / 2);
    pDC->DrawText(m_sides[side].m_sLabel, xStart, yStart);
  }
}

void TASTileType::DrawFlat3DTile(wxDC* pDC, wxPoint& pt, wxSize& sz)
{
  if (!m_bitmapFlat3D.IsOk()) {
    wxMemoryDC dcMem;
    if (!dcMem.IsOk()) {    //MJP: ERROR condition!
      return;
    }

    if (bHighQuality) {
      double scaleX = (double)sz.GetWidth() / (double)(g_tileSize * 3);
      double scaleY = (double)sz.GetHeight() / (double)(g_tileSize * 4);
      m_bitmapFlat3D.Create(sz.GetWidth(), sz.GetHeight());
      dcMem.SetUserScale(scaleX, scaleY);
    } else {
      m_bitmapFlat3D.Create((g_tileSize * 3), (g_tileSize * 4));
    }
    dcMem.SelectObject(m_bitmapFlat3D);

    wxPoint pt(0, 0);

    m_brush.SetColour(g_pContext->m_colSpaceBkgrnd);
    m_pen.SetColour(g_pContext->m_colSpaceBkgrnd);
    dcMem.SetBrush(m_brush);
    m_pen.SetStyle(wxSOLID);
    dcMem.SetPen(m_pen);

    dcMem.DrawRectangle(pt.x, pt.y, (g_tileSize * 3), (g_tileSize * 4));

    if (m_colTile == wxNullColour) {
      m_brush.SetColour(g_pContext->m_colDefaultTile);
    } else {
      m_brush.SetColour(m_colTile);
    }
    dcMem.SetBrush(m_brush);

    dcMem.SetFont(fontGlues3D);
    if (m_colText == wxNullColour) {
      m_pen.SetColour(g_pContext->m_colDefaultText);
      dcMem.SetTextForeground(g_pContext->m_colDefaultText);
    } else {
      m_pen.SetColour(m_colText);
      dcMem.SetTextForeground(m_colText);
    }
    dcMem.SetPen(m_pen);

    wxString s(_T("Name:"));
    wxCoord w, h;
    dcMem.DrawText(s, 3, 3);
    dcMem.GetTextExtent(s, &w, &h);
    dcMem.DrawText(m_sName, 3, h + 8);

    s = _T("Label:");
    dcMem.DrawText(s, (2 * g_tileSize) + 3, 3);
    dcMem.DrawText(m_sLabel, (2 * g_tileSize) + 3, h + 8);

    for (int dir=TAS_NORTH; dir<=TAS_DOWN; dir++) {
      DrawFlat3DSide(&dcMem, pt, g_tileSize, dir);
    }
  }

  if (m_bitmapFlat3D.IsOk() && pDC) {
    pDC->DrawBitmap(m_bitmapFlat3D, pt.x, pt.y, false);
  }
}

void TASTileType::DrawTile(wxDC* pDC, wxPoint& pt, int dim, unsigned displayMask)
{
  if (m_vecBitmaps.size() <= dim) {
    m_vecBitmaps.resize(dim+1);
  }

  if (!m_vecBitmaps[dim].m_bitmap.IsOk() || (m_vecBitmaps[dim].m_displayMask != displayMask)) {

    if (!m_image.IsOk() && m_sImagePath.Length() > 0) {
      m_image = wxImage(m_sImagePath);
    }

    if (m_image.IsOk()) {   //  This tiletype has an image designated by the input file
      m_vecBitmaps[dim].m_bitmap = wxBitmap(m_image.Scale(dim, dim));
      m_vecBitmaps[dim].m_displayMask = displayMask;
    } else {
      m_vecBitmaps[dim].m_displayMask = displayMask;

      wxMemoryDC dcMem;
      if (bHighQuality) {
        double scale = (double)dim / (double)g_tileSize;
        m_vecBitmaps[dim].m_bitmap.Create(dim, dim);
        dcMem.SelectObject(m_vecBitmaps[dim].m_bitmap);
        dcMem.SetUserScale(scale, scale);
      } else {
        m_vecBitmaps[dim].m_bitmap.Create(g_tileSize, g_tileSize);
        dcMem.SelectObject(m_vecBitmaps[dim].m_bitmap);
      }

      if (!dcMem.IsOk()) {    //MJP: ERROR condition!
        return;
      }

      wxPoint pt(0, 0);

      if (m_colTile == wxNullColour) {
        m_brush.SetColour(g_pContext->m_colDefaultTile);
        m_pen.SetColour(g_pContext->m_colDefaultTile);
      } else {
        m_brush.SetColour(m_colTile);
        m_pen.SetColour(m_colTile);
      }

      dcMem.SetBrush(m_brush);
      m_pen.SetStyle(wxSOLID);
      dcMem.SetPen(m_pen);
      dcMem.DrawRectangle(pt.x, pt.y, g_tileSize, g_tileSize);

      m_pen.SetColour(g_pContext->m_colDefaultText);
      dcMem.SetPen(m_pen);

      if (TAS_MASK_ISON(displayMask, TAS_MASK_DISPGLUES) || TAS_MASK_ISON(displayMask, TAS_MASK_DISPSTRENGTHS)) {
        m_pen.SetWidth(3);
        if (m_colText == wxNullColour) {
          m_pen.SetColour(g_pContext->m_colDefaultText);
          dcMem.SetTextForeground(g_pContext->m_colDefaultText);
        } else {
          m_pen.SetColour(m_colText);
          dcMem.SetTextForeground(m_colText);
        }

        dcMem.SetFont(fontGlues);
        for (int i=TAS_NORTH; i<TAS_UP; i++) {
          DrawBond(&dcMem, pt, i, TAS_MASK_ISON(displayMask, TAS_MASK_DISPGLUES));
        }

        wxCoord w, h;
        if (TAS_MASK_ISON(displayMask, TAS_MASK_DISPLABEL)) {
          dcMem.SetFont(fontLabel);
          dcMem.GetTextExtent(m_sLabel, &w, &h);
          dcMem.DrawText(m_sLabel, pt.x + (50-(w/2)), pt.y + 40);
        }
        dcMem.SetFont(*wxSWISS_FONT);
        if (TAS_MASK_ISON(displayMask, TAS_MASK_DISPNAME)) {
          dcMem.GetTextExtent(m_sName, &w, &h);
          dcMem.DrawText(m_sName, pt.x + (50-(w/2)), pt.y + 57);
        }
      }

      //  Scale the bitmap appropriately
      if (!bHighQuality && (m_vecBitmaps[dim].m_bitmap.GetHeight() != dim)) {
        wxImage image = m_vecBitmaps[dim].m_bitmap.ConvertToImage();
        m_vecBitmaps[dim].m_bitmap = wxBitmap(image.Scale(dim, dim));
      }
    }
  }

  if (m_vecBitmaps[dim].m_bitmap.IsOk() && pDC) {
    pDC->DrawBitmap(m_vecBitmaps[dim].m_bitmap, pt.x, pt.y, false);
  }
}

void TASTileType::DrawBond(wxDC* pDC, wxPoint& pt, int dir, bool bShowGlues, bool bSVG)
{
  int b = 2;
  int e = g_tileSize - b;
  int g = 5;
  int tg = 10;

  int strength = m_sides[dir].m_iStrength;

  wxCoord w, h;
  w = h = 0;

  if (bShowGlues) {
    pDC->GetTextExtent(m_sides[dir].m_sLabel, &w, &h);
    w = w/2;
  }

  if (strength > 0) {
    m_pen.SetStyle(wxSOLID);
    pDC->SetPen(m_pen);
  } else {
    m_pen.SetStyle(wxDOT);
    pDC->SetPen(m_pen);
  }

  if (dir == TAS_NORTH) {
    pDC->DrawLine(pt.x + b, pt.y + b, pt.x + e, pt.y + b);
    if (strength >= 2) {
      pDC->DrawLine(pt.x + b, pt.y + b+g, pt.x + e, pt.y + b+g);
      if (strength > 2) {
        pDC->DrawLine(pt.x + b, pt.y + b+(g/2), pt.x + e, pt.y + b+(g/2));
      }
    }
    if (strength > 0 && bShowGlues) {
      pDC->DrawText(m_sides[dir].m_sLabel, pt.x + ((e-b)/2)-w, pt.y + b+h-4);
    }
  } else if (dir == TAS_EAST) {
    pDC->DrawLine(pt.x + e, pt.y + b, pt.x + e, pt.y + e);
    if (strength >= 2) {
      pDC->DrawLine(pt.x + e-g, pt.y + b, pt.x + e-g, pt.y + e);
      if (strength > 2) {
        pDC->DrawLine(pt.x + e-(g/2), pt.y + b, pt.x + e-(g/2), pt.y + e);
      }
    }
    if (strength > 0 && bShowGlues) {
      wxCoord cX = pt.x + (g_tileSize - 10);
      wxCoord cY = pt.y + (g_tileSize/2) - w;
      pDC->DrawRotatedText(m_sides[dir].m_sLabel, cX, cY, 270.0);//pt.x + ((e-b)/2)-w, pt.y -e+h-3, 270.0);
    }
  } else if (dir == TAS_SOUTH) {
    pDC->DrawLine(pt.x + b, pt.y + e, pt.x + e, pt.y + e);
    if (strength >= 2) {
      pDC->DrawLine(pt.x + b, pt.y + e-g, pt.x + e, pt.y + e-g);
      if (strength > 2) {
        pDC->DrawLine(pt.x + b, pt.y + e-(g/2), pt.x + e, pt.y + e-(g/2));
      }
    }
    if (strength > 0 && bShowGlues) {
      pDC->DrawText(m_sides[dir].m_sLabel, pt.x + ((e-b)/2)-w, pt.y + e-tg-h);
    }
  } else if (dir == TAS_WEST) {
    pDC->DrawLine(pt.x + b, pt.y + b, pt.x + b, pt.y + e);
    if (strength >= 2) {
      pDC->DrawLine(pt.x + b+g, pt.y + b, pt.x + b+g, pt.y + e);
      if (strength > 2) {
        pDC->DrawLine(pt.x + b+(g/2), pt.y + b, pt.x + b+(g/2), pt.y + e);
      }
    }
    if (strength > 0 && bShowGlues) {
      wxCoord cX = pt.x + h;
      wxCoord cY = pt.y + (g_tileSize/2) + w;
      pDC->DrawRotatedText(m_sides[dir].m_sLabel, cX, cY, 90.0);//pt.x + (b-e)/2-(w+4), pt.y + h, 90.0);//((e-b)/2)-w, -h-tg-2);
    }
  }
}

void TASTileType::DrawSVG(wxDC* pDC, wxPoint& pt, int dim, unsigned displayMask)
{
  if (!m_image.IsOk() && m_sImagePath.Length() > 0) {
    m_image = wxImage(m_sImagePath);
  }

  if (m_image.IsOk()) {   //  This tiletype has an image designated by the input file
//MJP:  not handled
  } else {
    if (!pDC->IsOk()) {    //MJP: ERROR condition!
      return;
    }

/*    if (bHighQuality) {
      double scale = (double)dim / (double)g_tileSize;
      pDC->SetUserScale(scale, scale);
    }*/

    if (m_colTile == wxNullColour) {
      m_brush.SetColour(g_pContext->m_colDefaultTile);
      m_pen.SetColour(g_pContext->m_colDefaultTile);
    } else {
      m_brush.SetColour(m_colTile);
      m_pen.SetColour(m_colTile);
    }

    pDC->SetBrush(m_brush);
    m_pen.SetStyle(wxSOLID);
    pDC->SetPen(m_pen);
    pDC->DrawRectangle(pt.x, pt.y, g_tileSize, g_tileSize);

    m_pen.SetColour(g_pContext->m_colDefaultText);
    pDC->SetPen(m_pen);

    if (TAS_MASK_ISON(displayMask, TAS_MASK_DISPGLUES) || TAS_MASK_ISON(displayMask, TAS_MASK_DISPSTRENGTHS)) {
//      m_pen.SetWidth(3);
      if (m_colText == wxNullColour) {
        m_pen.SetColour(g_pContext->m_colDefaultText);
        pDC->SetTextForeground(g_pContext->m_colDefaultText);
      } else {
        m_pen.SetColour(m_colText);
        pDC->SetTextForeground(m_colText);
      }

      pDC->SetFont(fontGluesSVG);
      for (int i=TAS_NORTH; i<TAS_UP; i++) {
        DrawBond(pDC, pt, i, TAS_MASK_ISON(displayMask, TAS_MASK_DISPGLUES));
      }

      wxCoord w, h;
      if (TAS_MASK_ISON(displayMask, TAS_MASK_DISPLABEL)) {
        pDC->SetFont(fontLabelSVG);
        pDC->GetTextExtent(m_sLabel, &w, &h);
        pDC->DrawText(m_sLabel, pt.x + (50-(w/2)), pt.y + 40);
      }
      pDC->SetFont(*wxSWISS_FONT);
      if (TAS_MASK_ISON(displayMask, TAS_MASK_DISPNAME)) {
        pDC->GetTextExtent(m_sName, &w, &h);
        pDC->DrawText(m_sName, pt.x + (50-(w/2)), pt.y + 57);
      }
    }
  }
}

//  An example tile type definition to be read in from inFile:
//LABEL Y
//TILENAME Y2
//EASTBIND 0
//WESTBIND 0
//SOUTHBIND 2
//NORTHBIND 2
//UPBIND 0
//DOWNBIND 0
//WESTLABEL -99
//EASTLABEL -99
//NORTHLABEL y2
//SOUTHLABEL y1
//UPLABEL -99
//DOWNLABEL -99
//TILECOLOR 5693267
//TEXTCOLOR 4103935
//CREATE

TASFrontierLocation* TASFrontierSet::GetElement(size_t ind)
{
  if (ind >= size()) {
    return NULL;
  } else if (ind <= (size()/2)) {
    TASFrontierSet::iterator it = begin();
    while (ind) {
      it++;
      ind--;
    }
    return (TASFrontierLocation*)&(*it);
  } else {
    ind = (size() - ind) - 1;
    TASFrontierSet::reverse_iterator it = rbegin();
    while (ind) {
      it++;
      ind--;
    }
    return (TASFrontierLocation*)&(*it);
  }
}

bool TASFrontierSet::RemoveElement(const TASCoords& coords)
{
  TASCoords c(coords);
  TASFrontierLocation tfl(c);
  TASFrontierSet::iterator it = find(tfl);
  if (it != end()) {
    erase(it);
    return true;
  }
  return false;
}

bool TASTileSpace::LoadSpaceDefinitionFromFile(wxFFile* inFile)
{
  wxString fileContents;
  if (inFile->IsOpened() && inFile->ReadAll(&fileContents)) {
    return LoadSpaceDefinitionFromString(fileContents);
  }
  return false;
}

bool TASTileSpace::LoadSpaceDefinitionFromString(wxString sSeedDefn)
{
  ResetSpace();
  m_seed.m_iAvgX = 0;
  m_seed.m_iAvgY = 0;
  m_seed.m_iAvgZ = 0;
  bool bFirst = true;
  TASCoords coordsZero;
  m_seed.m_coordsLeftmost = coordsZero;

  int sumX, sumY, sumZ, numX, numY, numZ;
  sumX = sumY = sumZ = numX = numY = numZ = 0;

  TASCoords coordsMin(0, 0, 0);
  TASCoords coordsMax(0, 0, 0);

  wxString strTtypeName;

  wxString sLine = sSeedDefn.BeforeFirst('\n');
  sSeedDefn = sSeedDefn.AfterFirst('\n');

  m_seed.clear();

  wxString sMsg = _T("=============================================\n");
  sMsg += _T("Attempting load of the following tile assembly:\n");
  sMsg += sSeedDefn;
  sMsg += _T("\n");
  sMsg += _T("=============================================\n");
  g_pTASApp->SetSimDebugText(sMsg, true, false);

  bool bFailedMsgBox  = false;
  bool bFoundTemp     = false;

  while (sSeedDefn.Len() > 0 || sLine.Len() > 0) {
    wxStringTokenizer tok(sLine);
    wxString sTileName;

    long lX, lY, lZ;
    lX = lY = lZ = 0;

    if (tok.CountTokens() == 3) {           //  2-D tile
      sTileName = tok.GetNextToken();
      if (!tok.GetNextToken().ToLong(&lX) ||
          !tok.GetNextToken().ToLong(&lY))
      {
        wxString sMsg;
        if (!bFailedMsgBox) {
          sMsg  = _T("The assembly could not be correctly loaded. Please see\n");
          sMsg += _T("the Output (and Debug) message pane(s) for details.");
          wxMessageBox(sMsg, _T("Error loading assembly"), wxICON_ERROR);
          bFailedMsgBox = true;
        }
        sMsg = wxString::Format(_T("Load of assembly failed:  The following line contained an invalid coordinate for the tile location:\n%s\n"), sLine.c_str());
        g_pTASApp->SetSimOutputText(sMsg, true, false);
      } else {
        TASCoords c(lX, lY, lZ);
        TASSeedEntry se(sTileName, c);
        m_seed.push_back(se);

        numX++;
        numY++;
        sumX += (int)lX;
        sumY += (int)lY;
        if ((int)lX < coordsMin.x) coordsMin.x = (int)lX;
        if ((int)lY < coordsMin.y) coordsMin.y = (int)lY;
        if (coordsMax.x < (int)lX) coordsMax.x = (int)lX;
        if (coordsMax.y < (int)lY) coordsMax.y = (int)lY;

        if (bFirst) {
          m_seed.m_coordsLeftmost = c;
          bFirst = false;
        } else if (c.x < m_seed.m_coordsLeftmost.x) {
          m_seed.m_coordsLeftmost = c;
        } else if ((c.x == m_seed.m_coordsLeftmost.x) && (c.y < m_seed.m_coordsLeftmost.y)) {
          m_seed.m_coordsLeftmost = c;
        }

      }
    } else if (tok.CountTokens() == 4) {    //  3-D tile
      sTileName = tok.GetNextToken();
      if (!tok.GetNextToken().ToLong(&lX) ||
          !tok.GetNextToken().ToLong(&lY) ||
          !tok.GetNextToken().ToLong(&lZ))
      {
        wxString sMsg;
        if (!bFailedMsgBox) {
          sMsg  = _T("The assembly could not be correctly loaded. Please see\n");
          sMsg += _T("the Output (and Debug) message pane(s) for details.");
          wxMessageBox(sMsg, _T("Error loading assembly"), wxICON_ERROR);
          bFailedMsgBox = true;
        }
        sMsg = wxString::Format(_T("Load of assembly failed:  The following line contained an invalid coordinate for the tile location:\n%s\n"), sLine.c_str());
        g_pTASApp->SetSimOutputText(sMsg, true, false);
      } else {
        TASCoords c(lX, lY, lZ);
        TASSeedEntry se(sTileName, c);
        m_seed.push_back(se);

        numX++;
        numY++;
        numZ++;
        sumX += (int)lX;
        sumY += (int)lY;
        sumZ += (int)lZ;
        if ((int)lX < coordsMin.x) coordsMin.x = (int)lX;
        if ((int)lY < coordsMin.y) coordsMin.y = (int)lY;
        if ((int)lZ < coordsMin.z) coordsMin.z = (int)lZ;
        if (coordsMax.x < (int)lX) coordsMax.x = (int)lX;
        if (coordsMax.y < (int)lY) coordsMax.y = (int)lY;
        if (coordsMax.z < (int)lZ) coordsMax.z = (int)lZ;

        if (bFirst) {
          m_seed.m_coordsLeftmost = c;
          bFirst = false;
        } else if (c.x < m_seed.m_coordsLeftmost.x) {
          m_seed.m_coordsLeftmost = c;
        } else if ((c.x == m_seed.m_coordsLeftmost.x) && (c.y < m_seed.m_coordsLeftmost.y)) {
          m_seed.m_coordsLeftmost = c;
        }
      }
    } else {
      wxString sTempPrefix(_T("Temperature="));
      bool bError = false;
      if ((tok.CountTokens() == 1) && (sLine.StartsWith(sTempPrefix))) {
        wxString sTemp = sLine.SubString(sTempPrefix.Len(), sLine.Len());
        long lTemp = 0;
        if (sTemp.Contains(_T(";"))) {
          bFoundTemp = true;
          m_vectTemps.clear();
          m_currTemp = -1;
          wxArrayString arrStrTemps = wxStringTokenize(sTemp, _T(";"));
          if (arrStrTemps.Count() <= 0) {
            bFoundTemp = false;
            bError = true;
            wxString sErr = wxString::Format(_T("An invalid temperature value or sequence was specified: '%s'\n"), sLine.c_str());
            g_pTASApp->SetSimOutputText(sErr, true, false);
          }
          for (size_t i=0; i<arrStrTemps.Count(); i++) {
            wxString sCurr = arrStrTemps[i];
            long lTemp = 0;
            if (sCurr.IsNumber() && sCurr.ToLong(&lTemp)) {
              m_vectTemps.push_back((int)lTemp);
            } else {
              bFoundTemp = false;
              bError = true;
              wxString sErr = wxString::Format(_T("An invalid temperature value or sequence was specified: '%s'\n"), sLine.c_str());
              g_pTASApp->SetSimOutputText(sErr, true, false);
            }
          }
          if (bFoundTemp && !bError) {
            m_currTemp = 0;
            g_pTASApp->SetTemp(m_vectTemps[m_currTemp]);
          }
        } else if (sTemp.IsNumber() && sTemp.ToLong(&lTemp)) {
          bFoundTemp = true;
          g_pTASApp->SetTemp((int)lTemp);
        } else {
          bError = true;
        }
      }
      if (bError || (tok.CountTokens() != 1) || 
          ((!tok.GetNextToken().EndsWith(_T(".tds"))) && (!sLine.StartsWith(sTempPrefix))))
      {
        wxString sMsg;
        if (!bFailedMsgBox) {
          sMsg  = _T("The assembly could not be correctly loaded. Please see\n");
          sMsg += _T("the Output (and Debug) message pane(s) for details.");
          wxMessageBox(sMsg, _T("Error loading assembly"), wxICON_ERROR);
          bFailedMsgBox = true;
        }
        sMsg = wxString::Format(_T("Load of assembly failed:  The following line was invalid:\n%s\n\n"), sLine.c_str());
        g_pTASApp->SetSimOutputText(sMsg, true, false);
      }
    }

    sLine = sSeedDefn.BeforeFirst(_T('\n'));
    sSeedDefn = sSeedDefn.AfterFirst(_T('\n'));
  }

  if (!bFoundTemp) {        //  If a temperature is not explicitly specified, default to 2
    m_vectTemps.clear();
    m_currTemp = -1;
    g_pTASApp->SetTemp(2);
  }

  GrowSpace(coordsMin);
  GrowSpace(coordsMax);

  if (numX > 0) {
    m_seed.m_iAvgX = sumX / numX;
  }
  if (numY > 0) {
    m_seed.m_iAvgY = sumY / numY;
  }
  if (numZ > 0) {
    m_seed.m_iAvgZ = sumZ / numZ;
  }

  wxString sVal = ValidateSeed();
  if (sVal.Len() > 0) {
    wxString sMsg = wxString::Format(_T("Load of assembly failed:  No tile types matching the following names were found in the current tile set, so they were skipped: %s\n"), sVal.c_str());
    g_pTASApp->SetSimOutputText(sMsg, true, false);

    if (!bFailedMsgBox) {
      sMsg  = _T("The assembly could not be correctly loaded. Please see\n");
      sMsg += _T("the Output (and Debug) message pane(s) for details.");
      wxMessageBox(sMsg, _T("Error loading assembly"), wxICON_ERROR);
    }
  }
  LoadSeed();

  return true;
}

const int ID_TAS_SAVESPACEOPTIONS = 5001;

class TASSpaceSaveOptionsDlg : public wxDialog
{
public:
  TASSpaceSaveOptionsDlg(wxWindow* pWin);

  wxRadioBox* m_pRadioBox;
};

TASSpaceSaveOptionsDlg::TASSpaceSaveOptionsDlg(wxWindow* pWin)
                         : wxDialog(pWin, wxID_ANY, _T("Assembly output options"), wxDefaultPosition, wxDefaultSize)
{
  wxBoxSizer* pSizer = new wxBoxSizer(wxVERTICAL);

  wxString sText;
  sText += _T("Please select whether you want to save just the seed assembly\n");
  sText += _T("or the entire assembly as it currently exists.");

  wxStaticText* pText = new wxStaticText(this, wxID_ANY, sText);
  pSizer->Add(pText, 0, wxTOP | wxLEFT | wxRIGHT | wxBOTTOM, 10);

  wxString arrChoices[] = {_T("Seed only"), _T("Entire assembly") };
  m_pRadioBox = new wxRadioBox(this, ID_TAS_SAVESPACEOPTIONS, _T("Assembly to save"),
                               wxDefaultPosition, wxDefaultSize, 2, arrChoices);

  pSizer->Add(m_pRadioBox, 0, wxLEFT | wxRIGHT | wxBOTTOM, 10);

  wxBoxSizer* pSizerButtons = new wxBoxSizer(wxHORIZONTAL);
  wxButton* pOK = new wxButton(this, wxID_OK);
  pSizerButtons->Add(pOK);
  wxButton* pCancel = new wxButton(this, wxID_CANCEL);
  pSizerButtons->Add(pCancel, 0, wxLEFT | wxRIGHT, 10);
  pSizer->Add(pSizerButtons, 0, wxALIGN_RIGHT);
  pSizer->AddSpacer(5);

  SetAutoLayout(TRUE);
  SetSizer(pSizer);

  pSizer->SetSizeHints(this);
  pSizer->Fit(this);
}

bool TASTileSpace::SaveToFile(wxFFile* outFile)
{
  if (!outFile->IsOpened()) {
    return false;
  }

  //  See if the user wants to save the full current assembly
  //  or just the seed
  TASSpaceSaveOptionsDlg dlg(g_pTASApp->m_pSimFrame);
  if (dlg.ShowModal() == wxID_OK) {

    //  Write out the name of the current tileset file since
    //  this assembly is most likely associated with it
    outFile->Write(wxString::Format(_T("%s\n"), g_pContext->m_fnTilesetFile.GetFullName().c_str()));

    //  Write out the current temperature value, or the temperature sequence if there is one
    if (m_vectTemps.size() > 0) {
      wxString sTempSeq = _T("Temperature=");
      for (size_t i=0; i<m_vectTemps.size(); i++) {
        if (i > 0) {
          sTempSeq += _T(";");
        }
        sTempSeq += wxString::Format(_T("%d"), m_vectTemps[i]);
      }
      sTempSeq += _T("\n");
      outFile->Write(sTempSeq);
    } else {
      outFile->Write(wxString::Format(_T("Temperature=%d\n"), GetTemp()));
    }

    if (dlg.m_pRadioBox->GetSelection() == 0) {     //  Save only the seed
      SetSeed();
      for (size_t i=0; i<m_seed.size(); i++) {
        outFile->Write(wxString::Format(_T("%s %d %d %d\n"), 
                                        m_seed[i].m_sName.c_str(),
                                        m_seed[i].m_coords.x, m_seed[i].m_coords.y, m_seed[i].m_coords.z));
      }
    } else {  //  Save the entire current assembly

      for (size_t i=0; i<m_pLocationIndex->GetSize(); i++) {
        if ((m_pLocationIndex->GetLocation(i) != TAS_TILE_NOT_FOUND) &&
            (m_pLocationIndex->GetLocation(i) != TAS_TILE_FRONTIER)  &&
            (m_pLocationIndex->GetLocation(i) != TAS_TILE_FRONTIERDEAD))
        {
          TASCoords coords = CoordsFromIndex(i);

          outFile->Write(wxString::Format(_T("%s %d %d %d\n"), 
                                          m_tileset[m_pLocationIndex->GetLocation(i)]->m_sName.c_str(),
                                          coords.x, coords.y, coords.z));
        }
      }
    }
  }
  outFile->Close();

  return true;
}

size_t TASTileSpace::GetRandSeed()
{
  if (g_pTASApp) {
    return g_pTASApp->m_randSeed;
  } else {
    return 0;
  }
}

void TASTileSpace::SetRandSeed(unsigned long seed, bool bRandomSeed)
{
  if (g_pTASApp) {
    g_pTASApp->Srand(seed, bRandomSeed);
  }
}

void TASTileSpace::SetDimNum(int dimNum)
{
  if (m_dimNum != dimNum) {
    m_dimNum  = dimNum;
    m_numNbrs = (m_dimNum == 2) ? TAS_WEST : TAS_DOWN;
    g_pContext->m_iTASDimMode = (dimNum == 2 ? TAS_2D_MODE : TAS_3D_MODE);
    ResetSpace(true);
  }
}

void TASTileSpace::ResetSpace(bool bSaveSeed)
{
  if (!bSaveSeed) {
    m_seed.clear();
  }

  m_epoch         = 0;
  m_numTiles      = 0;
  m_bEpochCached  = false;

  if (m_vectTemps.size() > 0) {
    m_temp = m_vectTemps[0];
    m_currTemp = 0;
  }

  //  (Re)seed the random number generator
  SetRandSeed(g_pTASApp->m_randSeed);

  wxString sDeadLoc = wxString::Format(_T("Reseting tile assembly space.\n"));
  g_pTASApp->SetSimOutputText(sDeadLoc, true, false);

  if (m_dimNum == 2) {
    SetDims(TASCoords(-10, -10, 0), TASCoords(20, 20, 0));
  } else {
    //SetDims(TASCoords(-10, -10, -10), TASCoords(20, 20, 20));
    SetDims(TASCoords(-5, -5, -5), TASCoords(10, 10, 10));
  }

  LoadSeed();
}

TASSeed& TASSeed::operator=(TASSeed& tts)
{
  m_iAvgX           = tts.m_iAvgX;
  m_iAvgY           = tts.m_iAvgY;
  m_iAvgZ           = tts.m_iAvgZ;
  m_coordsLeftmost  = tts.m_coordsLeftmost;

  clear();
  for (size_t i=0; i<tts.size(); i++) {
    push_back(tts[i]);
  }

  return *this;
}

wxString TASTileSpace::ValidateSeed()
{
  wxString sRet;

  TASSeed seedTemp;
  seedTemp.m_iAvgX          = m_seed.m_iAvgX;
  seedTemp.m_iAvgY          = m_seed.m_iAvgY;
  seedTemp.m_iAvgZ          = m_seed.m_iAvgZ;
  seedTemp.m_coordsLeftmost = m_seed.m_coordsLeftmost;

  for (size_t i=0; i<m_seed.size(); i++) {
    if (m_tileset.IndexOf(m_seed[i].m_sName) == TAS_TILE_NOT_FOUND) {
      if (sRet.length() > 0) {
        sRet += _T(", ");
      }
      sRet += _T("'") + m_seed[i].m_sName + _T("'");
    } else {
      seedTemp.push_back(m_seed[i]);
    }
  }
  m_seed.clear();
  m_seed = seedTemp;

  return sRet;
}

bool TASTileSpace::LoadSeed()
{
  for (size_t i=0; i<m_seed.size(); i++) {
    if (m_tileset.IndexOf(m_seed[i].m_sName) != TAS_TILE_NOT_FOUND) {
      TASTileType* pTT = m_tileset.at(m_seed[i].m_sName);
      AddTile(m_seed[i].m_coords, pTT);
    } else {
      wxString sMsg = wxString::Format(_T("No tile type matching the name '%s' exists\nin the current tile set.  Unable to load seed."), m_seed[i].m_sName.c_str());
      wxMessageBox(sMsg, _T("Invalid Seed"), wxICON_ERROR);
      return false;
    }
  }
  m_numTiles = m_seed.size();

  //  Check to see if the seed is tau-stable and warn if it isn't
  bool bTauStable = true;
  if (m_seed.size() > 1) {
    for (size_t i=0; i<m_seed.size(); i++) {
      if (GetIncidentGlueStrengthSum(m_seed[i].m_coords) < m_temp) {
        bTauStable = false;
        break;
      }
    }
  }
  if (!bTauStable) {
    wxString sMsg = _T("Warning:  this seed assembly is not tau-stable!\n");
    g_pTASApp->SetSimOutputText(sMsg, true, false);
  }

  return true;
}

void TASTileSpace::SetSeed()
{
  m_seed.clear();
  for (size_t i=0; i<GetNumTransitionsToTypeInEpoch(0, TAS_BEGIN_TILES); i++) {
    TASLocationTransition* pTrans = GetTransitionToTypeInEpoch(0, i, TAS_BEGIN_TILES);
    TASTileType* pTT = m_tileset[pTrans->m_typeEnd];
    if (pTT->m_sName.Length() > 0) {
      TASSeedEntry se(pTT->m_sName, pTrans->m_coords);
      m_seed.push_back(se);
    }
  }
}

void TASTileSpace::ClearSeed()
{
  m_seed.clear();
}

size_t TASTileSpace::GetNumTransitionsInEpoch(size_t epoch)
{
  if (epoch >= m_listEpochCache.size()) {
    return 0;
  } else {
    return m_listEpochCache[epoch].m_listTransitions.GetCount();
  }
}

TASLocationTransition* TASTileSpace::GetTransitionInEpoch(size_t epoch, size_t changeNum)
{
  int numTransitions = GetNumTransitionsInEpoch(epoch);
  if ((numTransitions == 0) || (changeNum >= numTransitions)) {
    return NULL;
  }

  return &m_listEpochCache[epoch].m_listTransitions[changeNum];
}

size_t TASTileSpace::GetNumTransitionsToTypeInEpoch(size_t epoch, int endType)
{
  if (epoch >= m_listEpochCache.size()) {
    return 0;
  } else if (endType == TAS_BEGIN_EMPTY) {
    return (m_listEpochCache[epoch].m_listTransitions.GetCount() - m_listEpochCache[epoch].m_arrBases[endType]);
  } else {
    return (m_listEpochCache[epoch].m_arrBases[endType+1] - m_listEpochCache[epoch].m_arrBases[endType]);
  }
}

TASLocationTransition* TASTileSpace::GetTransitionToTypeInEpoch(size_t epoch, size_t changeNum, int endType)
{
  int numTransitions = GetNumTransitionsToTypeInEpoch(epoch, endType);
  if ((numTransitions == 0) || (changeNum >= numTransitions)) {
    return NULL;
  }

  return &m_listEpochCache[epoch].m_listTransitions[m_listEpochCache[epoch].m_arrBases[endType]+changeNum];
}

bool TASTileSpace::AddNeighborsToFrontierList(const TASCoords& coords)
{
  TASCoords coordsNbr;

  for (int i=TAS_NORTH; i<=m_numNbrs; i++) {
    //  Check for existence of neighbor in location list or frontier list, and add to frontier if necessary
    coordsNbr = m_nbrOffsets[i] + coords;
    size_t locType = m_pLocationIndex->GetLocation(Index(coordsNbr));
    if (((locType == TAS_TILE_NOT_FOUND) || (locType == TAS_TILE_FRONTIERDEAD)) &&
        (GetIncidentGlueStrengthSum(coordsNbr) >= m_temp))
    {
      if (locType == TAS_TILE_FRONTIERDEAD) {
        m_setFrontierDead.RemoveElement(coordsNbr);
      }
      TASFrontierLocation tfl(coordsNbr);
      m_setFrontier.insert(tfl);

      TASLocationTransition trans(coordsNbr, locType, TAS_TILE_FRONTIER);
      m_listEpochCache[m_epoch].AddTransition(trans);

      //  Make sure to recalculate the index from coordsNbr below, because the call to GetIncidentGlueStrengthSum
      //  above can trigger a growth of the space and therefore reindexing and a tricky bug!!!
      m_pLocationIndex->SetLocation(Index(coordsNbr), TAS_TILE_FRONTIER, m_epoch);
    }
  }

  return true;
}

bool TASTileSpace::FixFrontierInNeighborhood(const TASCoords& coords)
{
  int num = (m_dimNum == 2) ? 5 : 7;
  for (int i=0; i<num; i++) {
    TASCoords coordsCurr(coordsNbrhd[i] + coords);
    size_t ind = Index(coordsCurr);
    size_t ttn = m_pLocationIndex->GetLocation(ind);
    if ((ttn == TAS_TILE_NOT_FOUND) || (ttn == TAS_TILE_FRONTIER) || (ttn == TAS_TILE_FRONTIERDEAD) || m_pLocationIndex->GetEpoch(i) > m_epoch) {
      if (GetIncidentGlueStrengthSum(coordsCurr) >= m_temp) {
        if (ttn != TAS_TILE_FRONTIER) {           //  This location should be added to the frontier
          TASFrontierLocation tfl(coordsCurr);
          m_setFrontier.insert(tfl);

          TASLocationTransition trans(coordsCurr, ttn, TAS_TILE_FRONTIER);
          m_listEpochCache[m_epoch].AddTransition(trans);

          //  Make sure to recalculate the index from coordsNbr below, because the call to GetIncidentGlueStrengthSum
          //  above can trigger a growth of the space and therefore reindexing and a tricky bug!!!
          m_pLocationIndex->SetLocation(Index(coordsCurr), TAS_TILE_FRONTIER, m_epoch);
        }
      } else if ((ttn == TAS_TILE_FRONTIER) || (ttn == TAS_TILE_FRONTIERDEAD)) {      //  This location should be removed from the frontier

        TASFrontierLocation tfl(coordsCurr);
        TASFrontierSet::iterator iter;
        if (ttn == TAS_TILE_FRONTIER) {
          iter = m_setFrontier.find(tfl);
          if (iter != m_setFrontier.end()) {
            m_setFrontier.erase(iter);
          }
        } else {
          iter = m_setFrontierDead.find(tfl);
          if (iter != m_setFrontierDead.end()) {
            m_setFrontierDead.erase(iter);
          }
        }

        TASLocationTransition trans(coordsCurr, ttn, TAS_TILE_NOT_FOUND);
        m_listEpochCache[m_epoch].AddTransition(trans);

        m_pLocationIndex->SetLocation(Index(coordsCurr), TAS_TILE_NOT_FOUND, m_epoch);
      }
    }
  }
  return true;
}

bool TASTileSpace::AddTileToLocation(const TASCoords& coords)
{
  TASTileTypePtrVect* pTTVect = GetFittingTileTypes(coords);
  if (pTTVect) {
    size_t sizeList = pTTVect->size();
    if (sizeList == 1) {                             //  There is exactly one fitting tile type
      return AddTile(coords, pTTVect->at(0));
    } else if (sizeList > 1) {                       //  There is nondeterminism here

      if (g_pContext->m_iTASTileConcMode == TAS_TILECONC_MODE_ON) {       //  Utilize tile concentrations

        //  Find the sum of the concentrations of the fitting tile types
        int concSum = 0;
        for (size_t i=0; i<pTTVect->size(); i++) {
          concSum += pTTVect->at(i)->m_concentration;
        }

        //  Generate a random number between 0 and concSum
        int randVal = g_pTASApp->RandInRange(concSum);

        //  Choose the appropriate tile type based on randVal
        int currSum = 0;
        for (size_t j=0; j<pTTVect->size(); j++) {
          currSum += pTTVect->at(j)->m_concentration;
          if (randVal < currSum) {
            return AddTile(coords, pTTVect->at(j));
          }
        }
      } else {                                                              //  Randomly pick one of the fitting tiles
        if (g_pContext->m_bReportNondeterminism) {                              //  Report the nondeterminism
          wxString sNames;
          for (size_t n=0; n<pTTVect->size(); n++) {
            sNames += _T("'") + pTTVect->at(n)->m_sName + _T("'");
            if (n < (pTTVect->size()-1)) sNames += _T(", ");
          }
          wxString sNonDet = wxString::Format(_T("Nondeterminism detected:  %d possible tiletypes fit into location (%d,%d): %s\n"),
                                              pTTVect->size(), coords.x, coords.y, sNames.c_str());
          g_pTASApp->SetSimOutputText(sNonDet, true, false);
        }
        int index = g_pTASApp->RandInRange(sizeList);
        return AddTile(coords, pTTVect->at(index));
      }
    }
  }
  return false;
}

bool TASTileSpace::TileCanBind(TASTileType* pTtype, const TASCoords& coords)
{
  int strength = 0;

  for (int d2=TAS_NORTH; d2<=m_numNbrs; d2++) {
    TASCoords coordsNbr = m_nbrOffsets[d2] + coords;
    TASTileType* nbrTType = GetTileType(coordsNbr);

    if (nbrTType && (pTtype->m_sides[d2].m_hashNum == nbrTType->m_sides[TASOppositeDirection[d2]].m_hashNum))
    {
      strength += pTtype->m_sides[d2].m_iStrength;
      if (strength >= m_temp) {
        return true;
      }
    }
  }
  return false;
}

TASTileTypePtrVect* TASTileTypeSet::GetTileTypeListFromHash(TASTileSideHash& tsh, int temp)
{
  TASTileSideHashMap::iterator it = m_mapTileSides.find(tsh);
  if (it != m_mapTileSides.end()) {             //  There is an entry for tile side set, so it's been looked up before
    if (it->second.size() > 0) {                    //  There is at least one fitting tile type
      return &it->second;
    } else {                                        //  There are no fitting tile types
      return NULL;
    }
  } else {                                      //  There is no entry for this tile side set, so do a full search for matches
    TASTileTypePtrVect vectPTT;
    for (size_t i=0; i<size(); i++) {
      int strength = 0;
      for (int d2=TAS_NORTH; d2<=TAS_DOWN; d2++) {
        if (at(i)->m_sides[d2] == tsh.m_arr[d2])
        {
          strength += tsh.m_arr[d2].m_iStrength;
          if ((strength >= temp) && !g_pContext->m_bReport1stCondLD) {
            break;
          }
        }
      }
      if (strength >= temp) {
        if ((strength > temp) && g_pContext->m_bReport1stCondLD) {
//MJP: should pass coords in for better reporting!
          wxString sMsg = wxString::Format(_T("Tile type '%s' can bind with strength %d (greater than tau = %d), detected at step %d\n"), at(i)->m_sName.c_str(), strength, temp, g_pTileSpace->m_epoch);
          g_pTASApp->SetSimOutputText(sMsg, true);
        }
        vectPTT.push_back(at(i));
      }
    }

    m_mapTileSides[tsh] = vectPTT;                  //  Add an entry so this won't be calculated again
    if (vectPTT.size() > 0) {
      return &m_mapTileSides[tsh];
    } else {
      return NULL;
    }
  }
}

TASTileTypePtrVect* TASTileSpace::GetFittingTileTypes(const TASCoords& coords)
{
  TASCoords       coordsNbr;
  TASTileSideHash tsh;
  TASTileSide     tsEmpty;

  for (int d=TAS_NORTH; d<=m_numNbrs; d++) {
    coordsNbr = m_nbrOffsets[d] + coords;
    TASTileType* pTTNbr = GetTileType(coordsNbr);
    if (pTTNbr) {
      tsh.m_arr[d] = pTTNbr->m_sides[TASOppositeDirection[d]];
    } else {
      tsh.m_arr[d] = tsEmpty;
    }
  }
  tsh.CreateHash();
  return m_tileset.GetTileTypeListFromHash(tsh, m_temp);
}

bool TASTileSpace::SetTemp(int temp)
{
  if (temp > 0) {

    int origTemp  = m_temp;
    bool bTempUp  = (temp > m_temp);

    m_temp = temp;

    //  Need to clear the hash mapping fitting tile types to neighborhoods because it is based on temperature
    m_tileset.m_mapTileSides.clear();

    //  Always want to increase the epoch number by at least one for a temperature change,
    //  even if there were no changes to the assembly (as long as we're not in epoch 0)
    if (m_epoch > 0) {
      m_epoch++;
      m_listEpochCache.resize(m_epoch+1);
      m_listEpochCache[m_epoch] = TASTileEpoch(m_temp);
    }

    wxString sMsg = wxString::Format(_T("Temperature value (tau) set to %d\n"), m_temp);
    g_pTASApp->SetSimOutputText(sMsg, true, false);

    if (bTempUp) {                    //  Melt off tiles as necessary
      size_t numMeltedOff = 0;
      bool bDidMelt = true;

      while (bDidMelt) {
        size_t currNumMeltedOff = 0;
        bDidMelt = false;
        std::vector<TASCoords> vectCoordsToMelt;
        for (size_t i=0; i<m_pLocationIndex->GetSize(); i++) {
          size_t typeNum = m_pLocationIndex->GetLocation(i);
          if (TASIsTileType(typeNum)) {
            TASCoords c(CoordsFromIndex(i));
            if (!TileCanBind(m_tileset[typeNum], c)) {

              //  Don't melt off seed tiles
              bool bInSeed = false;
              for (size_t s=0; s<m_seed.size(); s++) {
                if (m_seed[s].m_coords == c) {
                  bInSeed = true;
                }
              }
              if (!bInSeed) {
                vectCoordsToMelt.push_back(c);
                bDidMelt = true;
              }
            }
          }
        }
        if (vectCoordsToMelt.size() > 0) {
          bDidMelt = true;
          for (size_t j=0; j<vectCoordsToMelt.size(); j++) {
            RemoveTile(vectCoordsToMelt[j]);
            currNumMeltedOff++;
          }
          numMeltedOff  += currNumMeltedOff;
          m_numTiles    -= currNumMeltedOff;

          wxString sMsg = wxString::Format(_T("In step %d, %d tiles melted off the assembly\n"), m_epoch, currNumMeltedOff);
          g_pTASApp->SetSimDebugText(sMsg, true, false);
        }
      }
      wxString sMsg = wxString::Format(_T("The temperature change from %d to %d caused %d tiles to melt off of the assembly\n"), origTemp, m_temp, numMeltedOff);
      g_pTASApp->SetSimOutputText(sMsg, true, false);
    } else {    //  The temperature went down, so clear and rebuild the frontier

      for (TASFrontierSet::iterator it = m_setFrontier.begin(); it != m_setFrontier.end(); it++) {
        m_pLocationIndex->SetLocation(Index(it->m_coords), TAS_TILE_NOT_FOUND, m_epoch);
      }
      m_setFrontier.clear();

      for (TASFrontierSet::iterator it = m_setFrontierDead.begin(); it != m_setFrontierDead.end(); it++) {
        m_pLocationIndex->SetLocation(Index(it->m_coords), TAS_TILE_NOT_FOUND, m_epoch);
      }
      m_setFrontierDead.clear();

      size_t t = 0;
      TASLocationTransition* pTrans = NULL;
      for (size_t e=0; e<=(size_t)m_epoch; e++) {
        t = 0;
        while (pTrans = GetTransitionToTypeInEpoch(e, t++, TAS_BEGIN_TILES)) {
          AddNeighborsToFrontierList(pTrans->m_coords);
        }
      }
    }

    return true;
  }
  return false;
}

bool TASTileSpace::AddMaxTilesToFrontier()
{
  return AddTileToFrontierHelper(false);
}

bool TASTileSpace::AddTileToFrontier()
{
  return AddTileToFrontierHelper(true);
}

bool TASTileSpace::KillFrontierLocation(TASFrontierLocation* pFrontLoc)
{
  bool bRet = false;

  if (!pFrontLoc) {
    return false;
  }

  TASCoords coords = pFrontLoc->m_coords;

  if (m_setFrontier.RemoveElement(coords)) {
    m_setFrontierDead.RemoveElement(coords);

    TASFrontierLocation tfl(coords);
    m_setFrontierDead.insert(tfl);

    bRet = true;

    size_t indDead = Index(coords);
    m_pLocationIndex->SetLocation(indDead, TAS_TILE_FRONTIERDEAD, m_epoch);

    if (m_listEpochCache.size() <= (size_t)m_epoch) {
      m_listEpochCache.resize(m_epoch+1);
      m_listEpochCache[m_epoch] = TASTileEpoch(m_temp);
    }

    TASLocationTransition trans(coords, TAS_TILE_FRONTIER, TAS_TILE_FRONTIERDEAD);
    m_listEpochCache[m_epoch].AddTransition(trans);

    wxString sDeadLoc = wxString::Format(_T("Location (%d,%d) being removed from frontier - no fitting tile type\n"), coords.x, coords.y);
    g_pTASApp->SetSimDebugText(sDeadLoc, true, false);
  }

  return bRet;
}

bool TASTileSpace::AddTileToFrontierHelper(bool bSingleTile)
{
  bool bAdded = false;

  //  For now, the size of the frontier space is limited to RAND_MAX (~= 2^15).
  //  This may need to be increased by generating a random number over a larger range if it
  //  is not sufficient.
  if (m_setFrontier.size() >= RAND_MAX) {
    wxMessageBox(_T("The frontier space is too large to be processed!"), _T("Frontier space exceeded limit"));
  } else if (m_setFrontier.size() == 0) {   //  There are no more frontier locations
    wxString sEmptyFrontier = wxString::Format(_T("No more tiles can be added:  the frontier is empty!\n"));
    g_pTASApp->SetSimOutputText(sEmptyFrontier, true, true);
  } else {
    bool bContinue = true;
    TASFrontierSet frontierOrig;

    if (bSingleTile) {
      bContinue = (!bAdded && (m_setFrontier.size() > 0));
    } else {
      frontierOrig = m_setFrontier;
      bContinue = true;
    }

    //  Randomly select a location from the frontier list
    int randInd = g_pTASApp->RandInRange(m_setFrontier.size());

    randInd--;  //  This is simply because the first line of the loop will increment it
    while (bContinue) {
      randInd = (randInd + 1) % m_setFrontier.size();

      if (!bSingleTile) {
//MJP: should this start with a random frontier location??
        for (TASFrontierSet::iterator it = frontierOrig.begin(); it != frontierOrig.end(); it++) {
          if (AddTileToLocation(it->m_coords)) {
            bAdded = true;
          } else {
            KillFrontierLocation((TASFrontierLocation*)&*it);
          }
        }
        bContinue = false;
      } else {
        TASFrontierLocation* frontLoc = m_setFrontier.GetElement(randInd);// *it;//*(it + randInd%m_setFrontier.size());// m_listFrontier.at((randInd+i)%m_listFrontier.size());
        if (AddTileToLocation(frontLoc->m_coords)) {  //  Increment iter by one, signifying a successful tile addition
          bAdded = true;
          break;
        } else {
          KillFrontierLocation(frontLoc);
        }
        bContinue = (!bAdded && (m_setFrontier.size() > 0));
      }
    }

    if (!bAdded) {
      size_t sizeFrontierTotal = m_setFrontier.size() + m_setFrontierDead.size();
      wxString sNoAdd = wxString::Format(_T("Stopping execution -> Unable to add tiles to any of the %d frontier locations\n"), sizeFrontierTotal);
      g_pTASApp->SetSimOutputText(sNoAdd, true, true);

      TASFrontierLocation tfl;
      for (TASFrontierSet::iterator it = m_setFrontier.begin(); it != m_setFrontier.end(); it++) {
        sNoAdd = wxString::Format(_T("  No fitting tile type for frontier location (%d,%d,%d)\n"), it->m_coords.x, it->m_coords.y, it->m_coords.z);
        g_pTASApp->SetSimDebugText(sNoAdd, true);
      }
      for (TASFrontierSet::iterator it = m_setFrontierDead.begin(); it != m_setFrontierDead.end(); it++) {
        sNoAdd = wxString::Format(_T("  No fitting tile type for frontier location (%d,%d,%d)\n"), it->m_coords.x, it->m_coords.y, it->m_coords.z);
        g_pTASApp->SetSimDebugText(sNoAdd, true);
      }
    }
  }

  //  Check if temperature programming is being used and if the temperature now needs to be adjusted
  if (!bAdded && (m_vectTemps.size() > 0) && (m_currTemp < (m_vectTemps.size()-1))) {
    wxString sTempChange = wxString::Format(_T("Switching to next temperature value\n"));
    g_pTASApp->SetSimOutputText(sTempChange, true, true);
    m_currTemp++;
    m_epoch--;  //  Reset the epoch count since it will be incremented by SetTemp
    SetTemp(m_vectTemps[m_currTemp]);
    bAdded = true;
  }

  return bAdded;
}

int TASTileSpace::GetIncidentGlueStrengthSum(const TASCoords& coords)
{
  int retSum = 0;
  TASTileType* pTT = NULL;
  
  //  Check for existence of north neighbor tile and get incident glue strength
  pTT = GetTileType(TASCoords(coords.x, coords.y + 1, coords.z));
  if (pTT != NULL) {
    retSum += pTT->m_sides[TAS_SOUTH].m_iStrength;
  }
  //  Check for existence of south neighbor tile and get incident glue strength
  pTT = GetTileType(TASCoords(coords.x, coords.y - 1, coords.z));
  if (pTT != NULL) {
    retSum += pTT->m_sides[TAS_NORTH].m_iStrength;
  }
  //  Check for existence of east neighbor tile and get incident glue strength
  pTT = GetTileType(TASCoords(coords.x + 1, coords.y, coords.z));
  if (pTT != NULL) {
    retSum += pTT->m_sides[TAS_WEST].m_iStrength;
  }
  //  Check for existence of west neighbor tile and get incident glue strength
  pTT = GetTileType(TASCoords(coords.x - 1, coords.y,  coords.z));
  if (pTT != NULL) {
    retSum += pTT->m_sides[TAS_EAST].m_iStrength;
  }
  if (m_dimNum > 2) {
    //  Check for existence of up neighbor tile and get incident glue strength
    pTT = GetTileType(TASCoords(coords.x, coords.y, coords.z + 1));
    if (pTT != NULL) {
      retSum += pTT->m_sides[TAS_DOWN].m_iStrength;
    }
    //  Check for existence of down neighbor tile and get incident glue strength
    pTT = GetTileType(TASCoords(coords.x, coords.y, coords.z - 1));
    if (pTT != NULL) {
      retSum += pTT->m_sides[TAS_UP].m_iStrength;
    }
  }

  return retSum;
}

TileLocationIndex::TileLocationIndex()
{
  m_growthFactor = 50000; //MJP: tune this
  m_pLocs = new std::vector<TileLocationIndexItem>;
  m_pLocs->resize(m_growthFactor, TileLocationIndexItem(TAS_TILE_NOT_FOUND, 0));
}

TileLocationIndex::~TileLocationIndex()
{
  if (m_pLocs) {
    delete m_pLocs;
  }
}

//MJP: can't just blindly resize!  Will need to transfer values...
size_t TileLocationIndex::GetLocation(size_t ind)
{
  if ((m_pLocs->size() <= ind) || (ind == TAS_TILE_NOT_FOUND)) {
    return TAS_TILE_NOT_FOUND;
  } else {
    return m_pLocs->at(ind).m_tiletype;
  }
}

size_t TileLocationIndex::GetEpoch(size_t ind)
{
  if ((m_pLocs->size() <= ind) || (ind == TAS_TILE_NOT_FOUND)) {
    return TAS_TILE_NOT_FOUND;
  } else {
    return m_pLocs->at(ind).m_epoch;
  }
}

void TileLocationIndex::SetLocation(size_t ind, size_t tileTypeNum, size_t epoch)
{
  while (m_pLocs->size() <= ind) {
    m_pLocs->resize(m_pLocs->size() + m_growthFactor, TileLocationIndexItem(TAS_TILE_NOT_FOUND, 0));
  }
  m_pLocs->at(ind).m_tiletype = tileTypeNum;
  m_pLocs->at(ind).m_epoch    = epoch;
}

///////////////////////
//// MJP: can the EpochListList actually just be a list of tileTypeNums???
////    no, (maybe) because then you can't rewind?
////      - or maybe make rewinding update
////          - no jumping around (should be okay)
////          - what about frontier?
///////////////////////

TASTileSpace::TASTileSpace(int dimNum)
{
  m_epoch             = 0;
  m_numTiles          = 0;
  m_dimNum            = dimNum;
  m_numNbrs           = (m_dimNum == 2) ? TAS_WEST : TAS_DOWN;
  m_dimGrowthRate     = TAS_SPACE_GROWTHINC_2D;
  m_pLocationIndex    = NULL;
  m_temp              = 2;

  m_nbrOffsets[0]     = TASCoords( 0,  1,  0);
  m_nbrOffsets[1]     = TASCoords( 1,  0,  0);
  m_nbrOffsets[2]     = TASCoords( 0, -1,  0);
  m_nbrOffsets[3]     = TASCoords(-1,  0,  0);
  m_nbrOffsets[4]     = TASCoords( 0,  0,  1);
  m_nbrOffsets[5]     = TASCoords( 0,  0, -1);

  SetRandSeed(0, true);

  ResetSpace();
}

TASTileSpace::~TASTileSpace()
{
  delete m_pLocationIndex;
}

size_t TASTileSpace::GetMaxEpoch()
{
  return m_listEpochCache.size();
}

void TASTileSpace::SetDims(const TASCoords& coordsMin, const TASCoords& coordsMax)
{
  m_coordsMin = coordsMin;
  m_coordsMax = coordsMax;

  m_coordsDims = TASCoords((m_coordsMax.x - m_coordsMin.x + 1),
                           (m_coordsMax.y - m_coordsMin.y + 1),
                           (m_coordsMax.z - m_coordsMin.z + 1));

  wxString dbgMsg = wxString::Format(_T("Setting tile space size to %s\n"), m_coordsDims.ToString().c_str());
  g_pTASApp->SetSimDebugText(dbgMsg, true);

  m_epoch = 0;

  delete m_pLocationIndex;
  m_pLocationIndex = new TileLocationIndex();

  m_listEpochCache.clear();
  m_setFrontier.clear();
  m_setFrontierDead.clear();
}

void TASTileSpace::GrowSpace(const TASCoords& coords)
{
  TASCoords coordsDimsOld(m_coordsDims);
  TASCoords coordsMinsOld(m_coordsMin);
  TASCoords coordsMaxesOld(m_coordsMax);

  if (g_pContext->m_iTASDimMode == TAS_2D_MODE) {
    m_dimGrowthRate = TAS_SPACE_GROWTHINC_2D;
  } else {
    m_dimGrowthRate = TAS_SPACE_GROWTHINC_3D;
  }

  if (coords.x < m_coordsMin.x) {
    m_coordsMin.x = coords.x - m_dimGrowthRate;
  }
  if (coords.x > m_coordsMax.x) {
    m_coordsMax.x = coords.x + m_dimGrowthRate;
  }
  if (coords.y < m_coordsMin.y) {
    m_coordsMin.y = coords.y - m_dimGrowthRate;
  }
  if (coords.y > m_coordsMax.y) {
    m_coordsMax.y = coords.y + m_dimGrowthRate;
  }
  if (g_pContext->m_iTASDimMode == TAS_3D_MODE) {
    if (coords.z < m_coordsMin.z) {
      m_coordsMin.z = coords.z - m_dimGrowthRate;
    }
    if (coords.z > m_coordsMax.z) {
      m_coordsMax.z = coords.z + m_dimGrowthRate;
    }
  }

  if ((coordsMinsOld == m_coordsMin) && (coordsMaxesOld == m_coordsMax)) {
    return;
  }

  m_coordsDims = TASCoords((m_coordsMax.x - m_coordsMin.x + 1),
                             (m_coordsMax.y - m_coordsMin.y + 1),
                             (m_coordsMax.z - m_coordsMin.z + 1));

  wxString dbgMsg = wxString::Format(_T("Growing space %s -> %s for coords %s\n"), coordsDimsOld.ToString().c_str(), m_coordsDims.ToString().c_str(), ((TASCoords)coords).ToString().c_str());
  g_pTASApp->SetSimDebugText(dbgMsg, true);

  TileLocationIndex* pNewLocationIndex = new TileLocationIndex();
  for (size_t i=0; i<m_pLocationIndex->GetSize(); i++) {
    size_t loc = m_pLocationIndex->GetLocation(i);
    if (loc != TAS_TILE_NOT_FOUND) {
      TASCoords tc = CoordsFromIndexHelper(i, coordsDimsOld, coordsMinsOld);
      size_t indNew = Index(tc);
      size_t ep  = m_pLocationIndex->GetEpoch(i);
      pNewLocationIndex->SetLocation(indNew, loc, ep);
    }
  }
  delete m_pLocationIndex;
  m_pLocationIndex = pNewLocationIndex;
}

size_t TASTileSpace::Index(const TASCoords& coords)
{
  while (coords.x < m_coordsMin.x || coords.y < m_coordsMin.y || coords.z < m_coordsMin.z ||
         coords.x > m_coordsMax.x || coords.y > m_coordsMax.y || coords.z > m_coordsMax.z)
  {
    GrowSpace(coords);
  }
  
  if (m_dimNum < 3) {
    return (m_coordsDims.x * (coords.y - m_coordsMin.y)) + (coords.x - m_coordsMin.x);
  } else {
    return (m_coordsDims.y * m_coordsDims.x * (coords.z - m_coordsMin.z)) +
           (m_coordsDims.x * (coords.y - m_coordsMin.y)) +
           (coords.x - m_coordsMin.x);
  }
}

TASCoords TASTileSpace::CoordsFromIndex(size_t ind)
{
  return CoordsFromIndexHelper(ind, m_coordsDims, m_coordsMin);
}

TASCoords TASTileSpace::CoordsFromIndexHelper(size_t ind, const TASCoords& dims, const TASCoords& mins)
{
  TASCoords retCoords;

  if (m_dimNum < 3) {
    retCoords.x = (ind % dims.x) + mins.x;
    retCoords.y = ((ind - retCoords.x) / dims.x) + mins.y;
    retCoords.z = 0;
  } else {
    retCoords.x = (ind % dims.x) + mins.x;
    ind = (ind - retCoords.x) / dims.x;
    retCoords.y = (ind % dims.y) + mins.y;
    retCoords.z = ((ind - retCoords.y) / dims.y) + mins.z;
  }

  return retCoords;
}

void TASTileSpace::TruncateEpochCache(size_t maxEpoch)
{
  while (m_listEpochCache.size() > (maxEpoch+1)) {
    m_listEpochCache.pop_back();
  }
}

TASTileType* TASTileSpace::GetTileType(const TASCoords& coords)
{
  if (!((TASCoords)coords).IsValid()) {
    return NULL;
  }

  size_t i = Index(coords);
  size_t ttn = m_pLocationIndex->GetLocation(i);
  if ((ttn == TAS_TILE_NOT_FOUND)     ||
      (ttn == TAS_TILE_FRONTIER)      ||
      (ttn == TAS_TILE_FRONTIERDEAD)/*  ||
      (m_pLocationIndex->GetEpoch(i) > epoch)*/)
  {
    return NULL;
  } else {
    return m_tileset[ttn];
  }
}

bool TASTileSpace::IsFrontierLocation(const TASCoords& coords)
{
  if (!((TASCoords)coords).IsValid()) {
    return false;
  }

  size_t i = Index(coords);
  size_t ttn = m_pLocationIndex->GetLocation(i);
  if ((ttn == TAS_TILE_FRONTIER) || (ttn == TAS_TILE_FRONTIERDEAD) /* && (m_pLocationIndex->GetEpoch(i) <= epoch)*/) {
    return true;
  } else {
    return false;
  }
}

//MJP: should optimize and not require both of these calls
bool TASTileSpace::IsDeadFrontierLocation(const TASCoords& coords)
{
  if (!((TASCoords)coords).IsValid()) {
    return false;
  }

  size_t i = Index(coords);
  size_t ttn = m_pLocationIndex->GetLocation(i);
  if (ttn == TAS_TILE_FRONTIERDEAD) {
    return true;
  } else {
    return false;
  }
}

size_t TASTileSpace::GetEpochAdded(const TASCoords& coords)
{
  size_t i = Index(coords);
  return m_pLocationIndex->GetEpoch(i);
}

bool TASTileSpace::AddTile(TASCoords coords, const TASTileType* pTT)
{
  size_t indexLoc = Index(coords);
  size_t locType = m_pLocationIndex->GetLocation(indexLoc);

  if ((locType != TAS_TILE_NOT_FOUND) && (locType != TAS_TILE_FRONTIER) && (locType != TAS_TILE_FRONTIERDEAD)) {  //  A tile already exists here
    return false;
  }

  size_t ind = m_tileset.IndexOf(pTT->m_sName);
  if (ind == TAS_TILE_NOT_FOUND) {
    return false;
  }

  if (m_listEpochCache.size() <= (size_t)m_epoch) {
    m_listEpochCache.resize(m_epoch+1);
    m_listEpochCache[m_epoch] = TASTileEpoch(m_temp);
  }

  TASLocationTransition trans(coords, locType, ind);
  m_listEpochCache[m_epoch].AddTransition(trans);

  m_pLocationIndex->SetLocation(indexLoc, ind, m_epoch);

  if (locType == TAS_TILE_FRONTIER) {
    m_setFrontier.RemoveElement(coords);
  } else if (locType == TAS_TILE_FRONTIERDEAD) {
    m_setFrontierDead.RemoveElement(coords);
  }

  AddNeighborsToFrontierList(coords);

//MJP: figure out how to make sure this is efficiently utilized
//  std::sort(m_listFrontier.begin(), m_listFrontier.end());

  TruncateEpochCache(m_epoch);

  return true;
}

bool TASTileSpace::RemoveTile(TASCoords coords)
{
  TASTileType* pTT = GetTileType(coords);
  if (!pTT) {      //  This location is already empty
    return true;
  }

  if ((size_t)m_epoch >= m_listEpochCache.size()) {
    m_listEpochCache.resize(m_epoch+1);
    m_listEpochCache[m_epoch] = TASTileEpoch(m_temp);
  }

  size_t ind = m_tileset.IndexOf(pTT->m_sName);
  if (ind == TAS_TILE_NOT_FOUND) {
    return false;
  }

  TASLocationTransition trans(coords, ind, TAS_TILE_NOT_FOUND);
  m_listEpochCache[m_epoch].AddTransition(trans);

  m_pLocationIndex->SetLocation(Index(coords), TAS_TILE_NOT_FOUND, m_epoch);

  FixFrontierInNeighborhood(coords);

  TruncateEpochCache(m_epoch);
  return true;
}

bool TASTileSpace::SetTileTypes(wxString& sTileTypeDefns)
{
  m_tileset = sTileTypeDefns;
  ResetSpace();
  return true;
}

bool TASTileSpace::SetTileTypes(TASTileTypeSet& setIn)
{
  m_tileset = setIn;
  ResetSpace();
/*
  for (size_t i=0; i<m_listEpochCache.size(); i++) {
    if (m_listEpochCache[i].m_listTilesChanged.size() > 0) {
      for (TASTileLocationList::iterator iter=m_listEpochCache[i].m_listTilesChanged.begin(); 
                                            iter!=m_listEpochCache[i].m_listTilesChanged.end(); iter++)
      {
        TASTileType ttOld = setTileTypesOld[iter->m_tileTypeNum];
        size_t ind = m_tileset.IndexOf(ttOld);
        if (ind != TAS_TILE_NOT_FOUND) {
         iter->m_tileTypeNum = ind;
        } else {
  //MJP: if a tiletype has been deleted, should I try to replace with an equivalent?
          m_listEpochCache[i].m_listTilesChanged.erase(iter);
        }
      }
    }
  }
*/
  return true;
}

bool TASTileSpace::IncrementEpoch()
{
  wxMutexLocker lock(*g_pMutexTileSpaceGuard);

  if (m_epoch == 0) {   //  If we're leaving epoch 0, save the current assembly as the seed
    SetSeed();
  }

  m_epoch++;
  m_bEpochCached = false;

  if (IsEpochCached(m_epoch)) {    //  Epoch was cached so now just fix the m_pLocationIndex and frontier sets
    m_bEpochCached = true;

    m_temp = m_listEpochCache[m_epoch].m_temp;

    for (size_t i=0; i<GetNumTransitionsInEpoch(m_epoch); i++) {
      TASLocationTransition* pTrans = GetTransitionInEpoch(m_epoch, i);
      if (pTrans) {
        size_t indexLoc = Index(pTrans->m_coords);

        m_pLocationIndex->SetLocation(indexLoc, pTrans->m_typeEnd, m_epoch);

        //  If this location ends in a frontier, add it to the frontier set
        if (pTrans->m_typeEnd == TAS_TILE_FRONTIER) {
          m_setFrontier.insert(TASFrontierLocation(pTrans->m_coords));
        } else if (pTrans->m_typeEnd == TAS_TILE_FRONTIERDEAD) {
          m_setFrontierDead.insert(TASFrontierLocation(pTrans->m_coords));
        }

        //  If this location started the transition in a frontier, remove it
        if (pTrans->m_typeBegin == TAS_TILE_FRONTIER) {
          m_setFrontier.RemoveElement(pTrans->m_coords);
        } else if (pTrans->m_typeBegin == TAS_TILE_FRONTIERDEAD) {
          m_setFrontierDead.RemoveElement(pTrans->m_coords);
        }
      }
    }
  } else if (g_pContext->m_bMaxTileAdditions) {
    if (!AddMaxTilesToFrontier()) {
      if (GetNumTransitionsInEpoch(m_epoch) == 0) {
        m_epoch--;
      }
      return false;
    }
  } else if (!AddTileToFrontier()) {
    if (GetNumTransitionsInEpoch(m_epoch) == 0) {
      m_epoch--;
    }
    return false;
  }

  for (size_t i=0; i<GetNumTransitionsInEpoch(m_epoch); i++) {
    TASLocationTransition* pTrans = GetTransitionInEpoch(m_epoch, i);
    if (TASIsTileType(pTrans->m_typeEnd) && !TASIsTileType(pTrans->m_typeBegin)) {
      m_numTiles++;
    } else if (TASIsTileType(pTrans->m_typeBegin) && !TASIsTileType(pTrans->m_typeEnd)) {
      m_numTiles--;
    }
  }

  return true;
}

bool TASTileSpace::DecrementEpoch()
{
  wxMutexLocker lock(*g_pMutexTileSpaceGuard);

  if (m_epoch > 0) {
    m_bEpochCached = true;

    m_temp = m_listEpochCache[m_epoch].m_temp;

    for (size_t i=0; i<GetNumTransitionsInEpoch(m_epoch); i++) {
      TASLocationTransition* pTrans = GetTransitionInEpoch(m_epoch, i);
      if (pTrans) {
        size_t indexLoc = Index(pTrans->m_coords);

        m_pLocationIndex->SetLocation(indexLoc, pTrans->m_typeBegin, m_epoch);

        //  If this location began in a frontier, add it to the frontier set
        if (pTrans->m_typeBegin == TAS_TILE_FRONTIER) {
          m_setFrontier.insert(TASFrontierLocation(pTrans->m_coords));
        } else if (pTrans->m_typeBegin == TAS_TILE_FRONTIERDEAD) {
          m_setFrontierDead.insert(TASFrontierLocation(pTrans->m_coords));
        }

        //  If this location ended the transition in a frontier, remove it
        if (pTrans->m_typeEnd == TAS_TILE_FRONTIER) {
          m_setFrontier.RemoveElement(pTrans->m_coords);
        } else if (pTrans->m_typeEnd == TAS_TILE_FRONTIERDEAD) {
          m_setFrontierDead.RemoveElement(pTrans->m_coords);
        }

        if (TASIsTileType(pTrans->m_typeEnd) && !TASIsTileType(pTrans->m_typeBegin)) {
          m_numTiles--;
        } else if (TASIsTileType(pTrans->m_typeBegin) && !TASIsTileType(pTrans->m_typeEnd)) {
          m_numTiles++;
        }
      }
    }

    m_epoch--;

    return true;
  } else {
    return false;
  }
}

int TASTileSpace::GetUnusedTileTypes(int epoch, TASTileTypeSet* pUnusedTypes)
{
  TASTileTypeSet usedTypes;

  size_t t = 0;

  for (size_t e=0; e<=(size_t)epoch; e++) {
    t = 0;
    for (size_t i=0; i<GetNumTransitionsToTypeInEpoch(e, TAS_BEGIN_TILES); i++) {
      //  This will automatically fail if the tiletype already exists in the list
      TASLocationTransition* pTrans = GetTransitionToTypeInEpoch(e, i, TAS_BEGIN_TILES);
      usedTypes.push_back(*m_tileset[pTrans->m_typeEnd], false);
    }
    //MJP: not worrying about removed tiles because they were at least used at some point
  }

  for (size_t i=0; i<m_tileset.size(); i++) {
    if (usedTypes.IndexOf(m_tileset[i]->m_sName) == TAS_TILE_NOT_FOUND) {
      pUnusedTypes->push_back(*m_tileset[i], false);
    }
  }

  usedTypes.clear();
  return pUnusedTypes->size();
}

void TASTileSpace::OutputChangesPerEpoch()
{
  for (int i=0; i<=m_epoch; i++) {
    OutputChangesInEpoch(i);
  }
}

void TASTileSpace::OutputChangesInEpoch(int epoch)
{
  wxString sMsg = wxString::Format(_T("== Epoch %d ==\nTemperature = %d\n"), epoch, m_listEpochCache[epoch].m_temp);
  g_pTASApp->SetSimDebugText(sMsg, true, false);

  size_t numTrans = GetNumTransitionsInEpoch(epoch);
  for (size_t i=0; i<numTrans; i++) {
    TASLocationTransition* pTrans = GetTransitionInEpoch(epoch, i);
    if (GetDimNum() == 2) {
      sMsg = wxString::Format(_T("(%d,%d)"), pTrans->m_coords.x, pTrans->m_coords.y);
    } else {
      sMsg = wxString::Format(_T("(%d,%d,%d)"), pTrans->m_coords.x, pTrans->m_coords.y, pTrans->m_coords.z);
    }
    sMsg += _T(" changed from ");
    if (TASIsTileType(pTrans->m_typeBegin)) {
      sMsg += wxString::Format(_T("tile type '%s'"), m_tileset.at(pTrans->m_typeBegin)->m_sName.c_str());
    } else if (pTrans->m_typeBegin == TAS_TILE_FRONTIERDEAD) {
      sMsg += _T("a dead frontier location");
    } else if (pTrans->m_typeBegin == TAS_TILE_FRONTIER) {
      sMsg += _T("a frontier location");
    } else if (pTrans->m_typeBegin == TAS_TILE_NOT_FOUND) {
      sMsg += _T("an empty location");
    }
    sMsg += _T(" to ");
    if (TASIsTileType(pTrans->m_typeEnd)) {
      sMsg += wxString::Format(_T("tile type '%s'"), m_tileset.at(pTrans->m_typeEnd)->m_sName.c_str());
    } else if (pTrans->m_typeEnd == TAS_TILE_FRONTIERDEAD) {
      sMsg += _T("a dead frontier location");
    } else if (pTrans->m_typeEnd == TAS_TILE_FRONTIER) {
      sMsg += _T("a frontier location");
    } else if (pTrans->m_typeEnd == TAS_TILE_NOT_FOUND) {
      sMsg += _T("an empty location");
    }
    sMsg += _T("\n");

    g_pTASApp->SetSimDebugText(sMsg, true, false);
  }
}

TASTileTypeVect::~TASTileTypeVect()
{
  clear();
}

void TASTileTypeVect::push_back(const TASTileType* pTT)
{
  std::vector<TASTileType*>::push_back(new TASTileType(*pTT));
}

void TASTileTypeVect::insert(TASTileTypeVect::iterator iter, const TASTileType* pTT)
{
  TASTileType* pNewTT = new TASTileType(*pTT);
  std::vector<TASTileType*>::insert(iter, pNewTT);
}

void TASTileTypeVect::erase(TASTileTypeVect::iterator iter)
{
  delete *iter;
  std::vector<TASTileType*>::erase(iter);
}

void TASTileTypeVect::clear()
{
  while (size() > 0) {
    delete at(size()-1);
    pop_back();
  }
}

void TASTileTypeSet::InvalidateBitmaps()
{
  for (size_t i=0; i<size(); i++) {
    at(i)->InvalidateBitmaps();
  }
}

bool TASTileTypeSet::SaveToFile(wxFFile* outFile)
{
  if (!outFile->IsOpened()) {
    return false;
  }

  TASTileTypeVect::iterator iter = m_vectTypes.begin();
  while (iter != m_vectTypes.end()) {
    outFile->Write((*iter)->Serialize());
    outFile->Write(wxString(_T("\n")));
    iter++;
  }

  outFile->Close();
  return true;
}

TASTileTypeSet& TASTileTypeSet::operator=(const wxString& sTTDefns)
{
  clear();
  size_t index  = 0;
  size_t len    = sTTDefns.Len();
  while (index < len) {
    TASTileType tt(sTTDefns, &index);
    if (tt.IsValid()) {
      push_back(tt, false);
    }
  }
  return *this;
}

TASTileTypeSetEx& TASTileTypeSetEx::operator=(const wxString& sTTDefns)
{
  TASTileTypeSet::operator =(sTTDefns);
  return *this;
}

TASTileTypeSet& TASTileTypeSet::operator=(TASTileTypeSet& tts)
{
  clear();
  for (size_t i=0; i<tts.size(); i++) {
    push_back(*tts[i], false);
  }
  return *this;
}

TASTileType* TASTileTypeSet::operator[](const size_t ind)
{
  return at(ind);
}

TASTileType* TASTileTypeSet::at(const size_t ind)
{
  if (ind >= m_vectTypes.size()) {
    return &g_EmptyTileType;
  }
  return m_vectTypes[ind];
}

TASTileType* TASTileTypeSet::operator[](const wxString& sName)
{
  return at(sName);
}

TASTileType* TASTileTypeSet::at(const wxString& sName)
{
  size_t ind = IndexOf(sName);
  if (ind == TAS_TILE_NOT_FOUND) {
    return &g_EmptyTileType;
  } else {
    return m_vectTypes[ind];
  }
}

bool TASTileTypeSet::push_back(const TASTileType& tt, bool bReIndex)
{
  if (IndexOf(tt.m_sName) != TAS_TILE_NOT_FOUND) {
    wxString sMsg = wxString::Format(_T("Error loading tile type named '%s' - a tile type with that name already exists!\n"), tt.m_sName.c_str());
    g_pTASApp->SetEditorOutputText(sMsg, true, true);

    return false;
  }

  m_vectTypes.push_back(&tt);
  m_mapNameIndex[tt.m_sName] = m_vectTypes.size() - 1;
  if (bReIndex) {
    ReIndex();
  } else {
    IndexTile(m_vectTypes[m_vectTypes.size()-1], m_vectTypes.size() - 1);
  }

  return true;
}

bool TASTileTypeSet::insert(const TASTileType& tt, size_t ind)
{
  if ((IndexOf(tt.m_sName) != TAS_TILE_NOT_FOUND) || (ind < 0) || (ind > size())) {
    wxString sMsg = wxString::Format(_T("Error inserting tile type named '%s' - a tile type with that name already exists or insertion is out of range!\n"), tt.m_sName.c_str());
    g_pTASApp->SetEditorOutputText(sMsg, true, true);

    return false;
  }

  size_t count = 0;
  TASTileTypeVect::iterator iter = m_vectTypes.begin();
  while ((count < ind) && iter != m_vectTypes.end()) {
    count++;
    iter++;
  }
  m_vectTypes.insert(iter, (TASTileType*)&tt);
  IndexTile(m_vectTypes[ind], ind);

  return true;
}

bool TASTileTypeSet::erase(size_t ind)
{
  if ((ind < 0) || (ind >= size())) {
    return false;
  }

  size_t count = 0;
  TASTileTypeVect::iterator iter = m_vectTypes.begin();
  while ((count < ind) && iter != m_vectTypes.end()) {
    count++;
    iter++;
  }
  m_vectTypes.erase(iter);
  ReIndex();

  return true;
}

void TASTileTypeSet::clear()
{
  m_vectTypes.clear();
  m_mapNameIndex.clear();
  m_mapGlueIndex.clear();
  m_mapTileSides.clear();
  ReIndex();
}

size_t TASTileTypeSet::size()
{
  return m_vectTypes.size();
}

size_t TASTileTypeSet::IndexOf(const TASTileType& tt)
{
  return IndexOf(tt.m_sName);
}

size_t TASTileTypeSet::IndexOf(const wxString& sName)
{
  std::map<wxString, size_t>::iterator iter = m_mapNameIndex.find(sName);
  if (iter == m_mapNameIndex.end()) {
    return TAS_TILE_NOT_FOUND;
  } else {
    return iter->second;
  }
}

void TASTileTypeSet::IndexTile(TASTileType* pTT, size_t num)
{
  if (IndexOf(pTT->m_sName) == TAS_TILE_NOT_FOUND) {
    m_mapNameIndex[pTT->m_sName] = num;
  }

  int numNbrs = (g_pContext->m_iTASDimMode == TAS_2D_MODE) ? TAS_WEST : TAS_DOWN;
  for (int j=TAS_NORTH; j<=numNbrs; j++) {
    wxString sGlueHash = wxString::Format(_T("%d:%s"), pTT->m_sides[j].m_iStrength, pTT->m_sides[j].m_sLabel.c_str());
    std::map<wxString, size_t>::iterator iter = m_mapGlueIndex.find(sGlueHash);
    if (iter == m_mapGlueIndex.end()) {
      m_mapGlueIndex[sGlueHash] = (num*6) + j;
    }
    pTT->m_sides[j].m_hashNum = m_mapGlueIndex[sGlueHash];
    std::map<size_t, TASTileTypePtrVect>::iterator iter2 = m_mmapGlues[j].find(m_mapGlueIndex[sGlueHash]);
    if (iter2 == m_mmapGlues[j].end()) {
      TASTileTypePtrVect v;
      v.push_back(pTT);
      m_mmapGlues[j][m_mapGlueIndex[sGlueHash]] = v;
    } else {
      m_mmapGlues[j][m_mapGlueIndex[sGlueHash]].push_back(pTT);
    }
  }
}

void TASTileTypeSetEx::IndexTile(TASTileType* pTT, size_t num)
{
  TASTileTypeSet::IndexTile(pTT, num);

  m_sArrNames.push_back(pTT->m_sName);

  if (m_sMapLabels.find(pTT->m_sLabel) != m_sMapLabels.end()) {
    m_sMapLabels[pTT->m_sLabel] = m_sMapLabels[pTT->m_sLabel] + 1;
  } else {
    m_sMapLabels[pTT->m_sLabel] = 1;
    m_sArrLabels.push_back(pTT->m_sLabel);
  }

  int numNbrs = (g_pContext->m_iTASDimMode == TAS_2D_MODE) ? TAS_WEST : TAS_DOWN;
  for (int i=TAS_NORTH; i<=numNbrs; i++) {
    if (m_sMapGlues.find(pTT->m_sides[i].m_sLabel) != m_sMapGlues.end()) {
      m_sMapGlues[pTT->m_sides[i].m_sLabel] = m_sMapGlues[pTT->m_sides[i].m_sLabel] + 1;
    } else {
      m_sMapGlues[pTT->m_sides[i].m_sLabel] = 1;
      m_sArrGlues.push_back(pTT->m_sides[i].m_sLabel);
    }
  }
}

void TASTileTypeSet::ReIndex()
{
  m_mapNameIndex.clear();
  m_mapGlueIndex.clear();
  m_mapTileSides.clear();
  for (int k=TAS_NORTH; k<TAS_DOWN; k++) {
    m_mmapGlues[k].clear();
  }

  for (size_t i=0; i<m_vectTypes.size(); i++) {
    IndexTile(m_vectTypes[i], i);
  }
}

void TASTileTypeSetEx::ReIndex()
{
  m_sArrNames.Clear();
  m_sArrLabels.Clear();
  m_sArrGlues.Clear();
  m_sMapLabels.clear();
  m_sMapGlues.clear();

  TASTileTypeSet::ReIndex();
}

TASTileType* TASTileTypeSet::GetEquivalentTileType(TASTileType* pTtype)
{
  int numNbrs = (g_pContext->m_iTASDimMode == TAS_2D_MODE) ? TAS_WEST : TAS_DOWN;
  for (int d=TAS_NORTH; d<=numNbrs; d++) {
    if (pTtype->m_sides[d].m_iStrength > 0) {
      std::vector<TASTileType*>* pVect = &m_mmapGlues[d][pTtype->m_sides[d].m_hashNum];
      for (size_t i=0; i<pVect->size(); i++) {
        if ((pVect->at(i)->m_sName != pTtype->m_sName) && (pVect->at(i)->IsEqual(pTtype, false))) {
          return pVect->at(i);
        }
      }
      return NULL;
    }
  }

  //  Should only get here for tiles with 0 strength glues on all sides
  for (size_t i=0; i<m_vectTypes.size(); i++) {
    if ((m_vectTypes[i]->m_sName != pTtype->m_sName) && (m_vectTypes[i]->IsEqual(pTtype, false))) {
      return m_vectTypes[i];
    }
  }
  return NULL;
}

