/////////////////////////////////////////////////////////////////////////////
// Name:        TASTiles.h
// Purpose:     Declares classes for tiles, tile space, etc.
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef __TASTiles__
#define __TASTiles__

#include "TAS.h"
#include <wx/ffile.h>
#include <wx/bitmap.h>
#include <wx/pen.h>
#include <wx/brush.h>
#include <wx/image.h>
#include <vector>
#include <map>
#include <set>


#define TAS_NORTH   0
#define TAS_EAST    1
#define TAS_SOUTH   2
#define TAS_WEST    3
#define TAS_UP      4
#define TAS_DOWN    5
extern int TASOppositeDirection[];
extern TASCoords coordsNbrhd[7];
wxString TASDirectionNumToString(int dirNum, bool bAllCaps = false);

#define TAS_EMPTY_TTYPE_STR         _T("<empty>")
#define TAS_FRONTIER_TTYPE_STR      _T("<frontier>")
#define TAS_FRONTIERDEAD_TTYPE_STR  _T("<dead frontier>")

#define TAS_SPACE_GROWTHINC_2D      200
#define TAS_SPACE_GROWTHINC_3D      40

#define TAS_TILE_NOT_FOUND          (size_t)(-1)
#define TAS_TILE_FRONTIER           (size_t)(-2)
#define TAS_TILE_FRONTIERDEAD       (size_t)(-3)
bool TASIsTileType(size_t t);

extern int g_tileSize;

class TASTileSide
{
public:
  TASTileSide() { m_iStrength = 0; m_sLabel = _T(""); m_hashNum = TAS_TILE_NOT_FOUND; };
  TASTileSide(int strength, wxString label) { m_iStrength = strength; m_sLabel = label; };
  bool operator==(const TASTileSide& right);
  bool operator!=(const TASTileSide& right) { return !operator==(right); };

  int       m_iStrength;
  wxString  m_sLabel;
  size_t    m_hashNum;
};

#define TAS_MASK_DISPLABEL      1
#define TAS_MASK_DISPNAME       2
#define TAS_MASK_DISPGLUES      4
#define TAS_MASK_DISPSTRENGTHS  8
#define TAS_MASK_DISPALL        (TAS_MASK_DISPLABEL | TAS_MASK_DISPNAME | TAS_MASK_DISPGLUES | TAS_MASK_DISPSTRENGTHS)

#define TAS_MASK_ISON(mask, dispobj) ((mask & dispobj) == dispobj)

class TASTileTypeBitmapHolder
{
public:
  TASTileTypeBitmapHolder() { m_displayMask = TAS_MASK_DISPALL; };

  wxBitmap  m_bitmap;
  unsigned  m_displayMask;
};

class TASTileType
{
public:
  TASTileType(int dims = 2);
  TASTileType(const TASTileType& tt);
  TASTileType(const wxString& sTTDefn,    //  This constructor takes a string representation of the tiletype
              size_t* index);             //  or multiple tiletypes and an index into the string denoting
                                          //  where the constructor should begin looking for the definintion,
                                          //  and updates that index value to point to the portion of the
                                          //  string following the current definition
  ~TASTileType();

  bool IsValid();
  bool IsEqual(TASTileType* pTtype, bool bCompareNameAndLabel = false, bool bAll = false);
  void SetEqual(const TASTileType& tt);
  
  bool Rotate(int degrees);         //  Rotation is done in the clockwise direction

  wxString Serialize();

  //  Functions for rendering tiles
  void DrawTile(wxDC* pDC, wxPoint& pt, int dim, unsigned displayMask);
  void DrawBond(wxDC* pDC, wxPoint& pt, int dir, bool bShowGlues, bool bSVG = false);
  void DrawFlat3DTile(wxDC* pDC, wxPoint& pt, wxSize& sz);
  void DrawFlat3DSide(wxDC* pDC, wxPoint& pt, int dim, int side);
  wxBitmap& GetBitmap(int dim);
  void InvalidateBitmaps();
  void DrawSVG(wxDC* pDC, wxPoint& pt, int dim, unsigned displayMask); 

  int            m_dims;
  int            m_concentration;
  wxString       m_sName;
  wxString       m_sLabel;
  wxString       m_sComment;
  wxString       m_sImagePath;
  TASTileSide    m_sides[6];
  wxImage        m_image;
  wxColour       m_colTile;
  wxColour       m_colText;
  wxBitmap       m_bitmapFlat3D;

protected:
  wxPen          m_pen;
  wxBrush        m_brush;
  std::vector<TASTileTypeBitmapHolder>  m_vecBitmaps;
};

typedef std::multimap<size_t, TASTileType*, std::less<size_t>, std::allocator<std::pair<const size_t, 
                                               TASTileType*> > > multimaptype;
typedef multimaptype::value_type multimapvaluetype;

//  TASTileTypeVect creates copies of the TASTileTypes that are to be inserted
class TASTileTypeVect : public std::vector<TASTileType*>
{
public:
  ~TASTileTypeVect();
  void push_back(const TASTileType* pTT);
  void insert(TASTileTypeVect::iterator iter, const TASTileType* pTT);
  void erase(TASTileTypeVect::iterator iter);
  void clear();
};

//  TASTileTypePtrVect is very simply just a vector of pointers to TASTileTypes
class TASTileTypePtrVect : public std::vector<TASTileType*>
{
public:
};

class TASTileSideHash
{
public:
  TASTileSideHash() {};
  TASTileSideHash(TASTileSide& tsNorth, TASTileSide& tsEast,
                  TASTileSide& tsSouth, TASTileSide& tsWest,
                  TASTileSide& tsUp,    TASTileSide& tsDown);

  void CreateHash();

  bool operator<(const TASTileSideHash& tshRight) const;
  bool operator==(const TASTileSideHash& tshRight);

  TASTileSide   m_arr[6];
  wxString      m_strHash;
};

class TASTileSideHashMap : public std::map<TASTileSideHash, TASTileTypePtrVect>
{

};


class TASTileTypeSet
{
public:
  TASTileTypeSet& operator=(const wxString& sTTDefns);
  TASTileTypeSet& operator=(TASTileTypeSet& tts);
  TASTileType* operator[](size_t ind);
  TASTileType* operator[](const wxString& sName);

  TASTileType* at(size_t ind);
  TASTileType* at(const wxString& sName);

  bool SaveToFile(wxFFile* outFile);

  size_t size();
  bool push_back(const TASTileType& tt, bool bReIndex = true);
  bool insert(const TASTileType& tt, size_t ind);
  bool erase(size_t ind);
  void clear();

  size_t IndexOf(const wxString& sName);
  size_t IndexOf(const TASTileType& tt);

  TASTileTypePtrVect* GetTileTypeListFromHash(TASTileSideHash& tsh, int temp);
  TASTileType* GetEquivalentTileType(TASTileType* pTtype);
  int GetUnusedTileTypes(int epoch, TASTileTypeSet* pUnusedTypes);

  //  This is used for detecting functionally equivalent tile types
  std::map<size_t, TASTileTypePtrVect> m_mmapGlues[6];

  TASTileSideHashMap m_mapTileSides;

  void InvalidateBitmaps();

  virtual void ReIndex();

protected:
  virtual void IndexTile(TASTileType* pTT, size_t num);

  TASTileTypeVect             m_vectTypes;
  std::map<wxString, size_t>  m_mapNameIndex;
  std::map<wxString, size_t>  m_mapGlueIndex;
};

class TASTileTypeSetEx : public TASTileTypeSet
{
public:
  TASTileTypeSetEx& operator=(const wxString& sTTDefns);

  void IndexTile(TASTileType* pTT, size_t num);
  void ReIndex();

  void SetDiskMod(bool bIsMod)  { m_bModifiedFromDisk = bIsMod; };
  void SetSimMod(bool bIsMod)   { m_bModifiedFromSim  = bIsMod; };
  bool GetDiskMod()             { return m_bModifiedFromDisk; };
  bool GetSimMod()              { return m_bModifiedFromSim; };

  wxArrayString                 m_sArrNames;
  wxArrayString                 m_sArrLabels;
  wxArrayString                 m_sArrGlues;
  std::map<wxString, int>       m_sMapLabels;
  std::map<wxString, int>       m_sMapGlues;

protected:
  bool                          m_bModifiedFromDisk;
  bool                          m_bModifiedFromSim;
};

class TASFrontierLocation
{
public:
  TASFrontierLocation() {};
  TASFrontierLocation(const TASCoords& coords) { m_coords = coords; };
  
  inline bool operator<(const TASFrontierLocation& tfl) { return (m_coords < tfl.m_coords); };
  inline bool operator<(const TASFrontierLocation& tfl) const { return (m_coords < tfl.m_coords); };

  TASCoords   m_coords;
};

class TASFrontierSet : public std::set<TASFrontierLocation>
{
public:
  TASFrontierLocation* GetElement(size_t ind);
  bool RemoveElement(const TASCoords& coords);
};

class TileLocationIndexItem
{
public:
  TileLocationIndexItem() { m_tiletype = TAS_TILE_NOT_FOUND; m_epoch = 0; };
  TileLocationIndexItem(size_t tiletype, size_t epoch) { m_tiletype = tiletype; m_epoch = epoch; };

  size_t  m_tiletype;
  size_t  m_epoch;
};

//  The TileLocationIndex contains a value for every possible tile location in the current space.
//  The value for each location is one of the following:
//    - for locations with a tile, the tileType number
//    - for locations without a tile but on the frontier, TAS_TILE_FRONTIER
//    - for locations on the frontier but for which there is not fitting tile type, TAS_TILE_FRONTIERDEAD
//    - for all other locations, TAS_TILE_NOT_FOUND
class TileLocationIndex
{
public:
  TileLocationIndex();
  ~TileLocationIndex();

  size_t GetSize() { return m_pLocs->size(); };
  size_t GetLocation(size_t ind);
  size_t GetEpoch(size_t ind);
  void SetLocation(size_t ind, size_t tileTypeNum, size_t epoch);

protected:
  std::vector<TileLocationIndexItem>*  m_pLocs;
  size_t                m_growthFactor;
};

//  An instance of the class TASLocationTransition keeps track of one
//  transition from one of the following types to another at a given locaiton:
//  0 - maxTileTypeNum (the number representing some tile type)
//  TAS_TILE_NOT_FOUND
//  TAS_TILE_FRONTIER
//  TAS_TILE_FRONTIERDEAD
class TASLocationTransition
{
public:
  TASLocationTransition() { m_typeBegin = TAS_TILE_NOT_FOUND; m_typeEnd = TAS_TILE_NOT_FOUND; };
  TASLocationTransition(const TASLocationTransition& trans) { m_coords = trans.m_coords; m_typeBegin = trans.m_typeBegin; m_typeEnd = trans.m_typeEnd; };
  TASLocationTransition(TASCoords& coords, size_t typeStart, size_t typeEnd) { m_coords = coords; m_typeBegin = typeStart; m_typeEnd = typeEnd; };

  bool operator==(const TASLocationTransition& trans) { return (m_coords == trans.m_coords && m_typeBegin == trans.m_typeBegin && m_typeEnd == trans.m_typeEnd); };
  TASLocationTransition& operator=(const TASLocationTransition& trans) { m_coords = trans.m_coords; m_typeBegin = trans.m_typeBegin; m_typeEnd = trans.m_typeEnd; return *this; };

  TASCoords       m_coords;
  size_t          m_typeBegin;
  size_t          m_typeEnd;
};
WX_DECLARE_OBJARRAY(TASLocationTransition, TASLocationTransitionList);

//  An instance of TASTileEpoch stores the set of transitions that occurred
//  in an epoch.  The transitions are stored in a single list which is
//  grouped by the types of the ends of the transitions, in the following order:
//  <tile types>
//  TAS_TILE_FRONTIERDEAD
//  TAS_TILE_FRONTIER
//  TAS_TILE_NOT_FOUND
enum base_types { TAS_BEGIN_TILES, TAS_BEGIN_DEADFRONTIER, TAS_BEGIN_FRONTIER, TAS_BEGIN_EMPTY };
class TASTileEpoch
{
public:
  TASTileEpoch();
  TASTileEpoch(int temp);

  void ConstructorHelper();
  bool AddTransition(TASLocationTransition& trans);
  bool RemoveTransition(TASLocationTransition& trans);
  //bool RemoveTransitionsToType(int baseType);
  bool RemoveTransitionAtLocation(TASCoords& coords);
  void ClearTransitions();
  size_t GetEnd();

  TASLocationTransitionList   m_listTransitions;
  int                         m_arrBases[4];
  int                         m_temp;
};
class TASTileEpochList : public std::vector<TASTileEpoch> {};

class TASSeedEntry
{
public:
  TASSeedEntry(wxString sName, TASCoords& coords) { m_sName = sName; m_coords = coords; };

  wxString    m_sName;
  TASCoords   m_coords;
};
class TASSeed : public std::vector<TASSeedEntry>
{
public:
  TASSeed() { m_iAvgX = 0; m_iAvgY = 0; };

  TASSeed& operator=(TASSeed& tts);

  int       m_iAvgX;
  int       m_iAvgY;
  int       m_iAvgZ;
  TASCoords m_coordsLeftmost;
};

class TASTileSpace : public TASSpace
{
public:
  TASTileSpace(int dimNum = 2);
  ~TASTileSpace();

  bool IncrementEpoch();
  bool DecrementEpoch();
  bool IsEpochCached(size_t epoch) { return m_listEpochCache.size() >= (epoch+1); };

  bool AddTile(TASCoords coords, const TASTileType* pTT);
  bool RemoveTile(TASCoords coords);

  bool SetTileTypes(wxString& sTileTypeDefns);
  bool SetTileTypes(TASTileTypeSet& setIn);

  int GetTemp() { return m_temp; };
  bool SetTemp(int temp);

  size_t Index(const TASCoords& coords);
  TASCoords CoordsFromIndex(size_t ind);

  int GetXDim() { return m_coordsDims.x; };
  int GetYDim() { return m_coordsDims.y; };
  int GetZDim() { return m_coordsDims.z; };

  int GetDimNum() { return m_dimNum; };
  void SetDimNum(int dimNum);

  size_t GetRandSeed();
  void   SetRandSeed(unsigned long seed, bool bRandomSeed = false);

  size_t GetMaxEpoch();

  size_t GetNumTransitionsInEpoch(size_t epoch);
  TASLocationTransition* GetTransitionInEpoch(size_t epoch, size_t changeNum);

  //  The endType parameter for the following two functions should be one of the following:
  //    TAS_BEGIN_TILES, TAS_BEGIN_DEADFRONTIER, TAS_BEGIN_FRONTIER, TAS_BEGIN_EMPTY
  size_t GetNumTransitionsToTypeInEpoch(size_t epoch, int endType);
  TASLocationTransition* GetTransitionToTypeInEpoch(size_t epoch, size_t changeNum, int endType);

  TASTileType* GetTileType(const TASCoords& coords);
  bool IsFrontierLocation(const TASCoords& coords);
  bool IsDeadFrontierLocation(const TASCoords& coords);
  bool KillFrontierLocation(TASFrontierLocation* pFrontLoc);
  size_t GetEpochAdded(const TASCoords& coords);

  int GetUnusedTileTypes(int epoch, TASTileTypeSet* pUnusedTypes);

//MJP: could optimize saving and loading by also saving frontier so tiles don't have to be added one at a time with full calculations
  bool LoadSpaceDefinitionFromFile(wxFFile* inFile);
  bool LoadSpaceDefinitionFromString(wxString sSeedDefn);
  bool SaveToFile(wxFFile* outFile);

  void ClearSeed();
  wxString ValidateSeed();                  //  If the seed is valid, the return string is empty.  Otherwise, it
                                            //  is the first tile type name not found in the current tile set
  bool LoadSeed();
  void ResetSpace(bool bSaveSeed = false);

  void OutputChangesPerEpoch();
  void OutputChangesInEpoch(int epoch);

  void TruncateEpochCache(size_t maxEpoch);

//MJP: this group of functions can probably be much more optimized
  TASTileTypePtrVect* GetFittingTileTypes(const TASCoords& coords);
  int GetIncidentGlueStrengthSum(const TASCoords& coords);
  bool AddTileToLocation(const TASCoords& coords);
  bool TileCanBind(TASTileType* pTtype, const TASCoords& coords);
  bool AddTileToFrontier();
  bool AddMaxTilesToFrontier();
  bool AddNeighborsToFrontierList(const TASCoords& coords);
  bool FixFrontierInNeighborhood(const TASCoords& coords);

  TASTileTypeSet        m_tileset;
  TASCoords             m_coordsMin;
  TASCoords             m_coordsMax;
  TASCoords             m_coordsDims;           //  Pre-calculated dimension sizes for efficiency
  TASFrontierSet        m_setFrontier;          //  Set of current frontier locations
  TASFrontierSet        m_setFrontierDead;      //  Set of current frontier locations where no tile type can attach
//MJP: change this to type size_t
  int                   m_epoch;
  size_t                m_numTiles;
  bool                  m_bEpochCached;
  TASSeed               m_seed;
  std::vector<int>      m_vectTemps;            //  If temperature programming is being used, this will
                                                //  be the list of temperatures to apply
  int                   m_currTemp;             //  The index into m_vectTemps of the temperature currently
                                                //  being used
  TileLocationIndex*    m_pLocationIndex;       //  Effectively an n-dimensional array to directly index all current space locations

protected:
  void GrowSpace(const TASCoords& coords);
  void SetDims(const TASCoords& coordsMin, const TASCoords& coordsMax);   //  Deletes existing space and sets the requested dimensions

  TASCoords CoordsFromIndexHelper(size_t ind, const TASCoords& dims, const TASCoords& mins);
  void SetSeed();                                 //  Sets the seed equal to all the tiles in epoch 0

  bool AddTileToFrontierHelper(bool bSingleTile);

  int                     m_dimNum;               //  Is space 2-D or 3-D (MJP: currently hardcoded to 2-D)
  int                     m_numNbrs;              //  The number of neighbors that a location has (depends on dimension setting)
  int                     m_temp;                 //  The temperature specifier for the number of valid bonds necessary for a tile to stick
  int                     m_dimGrowthRate;
  TASTileEpochList        m_listEpochCache;

  TASCoords               m_nbrOffsets[6];
};

#endif /* __TASTiles__ */

